from __future__ import annotations

import logging
import os
import re

import libsbml
import pandas as pd
from napistu import consensus
from napistu import constants
from napistu import identifiers
from napistu import sbml_dfs_utils
from napistu import source
from napistu import utils

from napistu.constants import BQB

from napistu.ingestion.constants import SBML_ANNOTATION_METHOD_GET_COMPARTMENT
from napistu.ingestion.constants import SBML_ANNOTATION_METHOD_GET_REACTION
from napistu.ingestion.constants import SBML_ANNOTATION_METHOD_GET_SPECIES
from napistu.ingestion.constants import SBML_COMPARTMENT_DICT_ID
from napistu.ingestion.constants import SBML_COMPARTMENT_DICT_IDENTIFIERS
from napistu.ingestion.constants import SBML_COMPARTMENT_DICT_NAME
from napistu.ingestion.constants import SBML_COMPARTMENT_DICT_SOURCE
from napistu.ingestion.constants import SBML_COMPARTMENTALIZED_SPECIES_DICT_NAME
from napistu.ingestion.constants import SBML_COMPARTMENTALIZED_SPECIES_DICT_SOURCE
from napistu.ingestion.constants import SBML_REACTION_ATTR_GET_GENE_PRODUCT
from napistu.ingestion.constants import SBML_SPECIES_DICT_ID
from napistu.ingestion.constants import SBML_SPECIES_DICT_IDENTIFIERS
from napistu.ingestion.constants import SBML_SPECIES_DICT_NAME
from napistu.ingestion.constants import SMBL_ERROR_CATEGORY
from napistu.ingestion.constants import SMBL_ERROR_DESCRIPTION
from napistu.ingestion.constants import SMBL_ERROR_MESSAGE
from napistu.ingestion.constants import SMBL_ERROR_NUMBER
from napistu.ingestion.constants import SMBL_ERROR_SEVERITY
from napistu.ingestion.constants import SMBL_REACTION_DICT_ID
from napistu.ingestion.constants import SMBL_REACTION_DICT_IDENTIFIERS
from napistu.ingestion.constants import SMBL_REACTION_DICT_IS_REVERSIBLE
from napistu.ingestion.constants import SMBL_REACTION_DICT_NAME
from napistu.ingestion.constants import SMBL_REACTION_DICT_SOURCE
from napistu.ingestion.constants import SMBL_REACTION_SPEC_RSC_ID
from napistu.ingestion.constants import SMBL_REACTION_SPEC_SBO_TERM
from napistu.ingestion.constants import SMBL_REACTION_SPEC_SC_ID
from napistu.ingestion.constants import SMBL_REACTION_SPEC_STOICHIOMETRY
from napistu.ingestion.constants import SMBL_SUMMARY_COMPARTMENTS
from napistu.ingestion.constants import SMBL_SUMMARY_N_REACTIONS
from napistu.ingestion.constants import SMBL_SUMMARY_N_SPECIES
from napistu.ingestion.constants import SMBL_SUMMARY_PATHWAY_ID
from napistu.ingestion.constants import SMBL_SUMMARY_PATHWAY_NAME

from fs import open_fs

logger = logging.getLogger(__name__)


class SBML:
    """
    System Biology Markup Language Connections.

    Attributes
    ----------
    document
        Connection to the SBML document
    model
        Connection to the SBML model

    Methods
    -------
    summary()
        Prints a summary of the sbml model
    sbml_errors(reduced_log, return_df)
        Print a summary of all errors in the SBML file

    """

    def __init__(
        self,
        sbml_path: str,
    ) -> None:
        """
        Connects to an SBML file

        Parameters
        ----------
        sbml_path : str
            path to a .sbml file.

        Returns
        -------
        None.
        """

        reader = libsbml.SBMLReader()
        if os.path.exists(sbml_path):
            self.document = reader.readSBML(sbml_path)
        else:
            with open_fs(os.path.dirname(sbml_path)) as fs:
                txt = fs.readtext(os.path.basename(sbml_path))
                self.document = reader.readSBMLFromString(txt)

        if self.document.getLevel() < 3:
            raise ValueError(
                f"SBML model is level {self.document.getLevel()}, only SBML 3 is supported"
            )

        self.model = self.document.getModel()

        # check for critical sbml errors
        errors = self.sbml_errors(reduced_log=False, return_df=True)
        if errors is not None:
            critical_errors = errors[errors[SMBL_ERROR_SEVERITY] >= 2]
            critical_errors = set(critical_errors[SMBL_ERROR_DESCRIPTION].unique())
            known_errors = {"<layout> must have 'id' and may have 'name'"}

            found_known_errors = known_errors.intersection(critical_errors)
            if len(found_known_errors) > 0:
                logger.warning(
                    f"The following known errors were found: {found_known_errors}"
                )

            unknown_critical_errors = critical_errors - known_errors
            if len(unknown_critical_errors) != 0:
                critical_errors = ", ".join(unknown_critical_errors)
                raise ValueError(
                    f"Critical errors were found when reading the sbml file: {critical_errors}"
                )

    def summary(self) -> pd.DataFrame:
        """Returns a pd.DataFrame summary of an SBML model."""
        model = self.model

        model_summaries = dict()

        model_summaries[SMBL_SUMMARY_PATHWAY_NAME] = model.getName()
        model_summaries[SMBL_SUMMARY_PATHWAY_ID] = model.getId()

        model_summaries[SMBL_SUMMARY_N_SPECIES] = model.getNumSpecies()
        model_summaries[SMBL_SUMMARY_N_REACTIONS] = model.getNumReactions()

        compartments = [
            model.getCompartment(i).getName() for i in range(model.getNumCompartments())
        ]
        compartments.sort()
        model_summaries[SMBL_SUMMARY_COMPARTMENTS] = ",\n".join(compartments)

        model_summaries_dat = pd.DataFrame(model_summaries, index=[0]).T

        return utils.style_df(model_summaries_dat)  # type: ignore

    def sbml_errors(self, reduced_log: bool = True, return_df: bool = False):
        """
        Format and print all SBML errors

        Parameters
        ----------
        reduced_log : bool
            Reduced log aggregates errors across categories an severity levels
        return_df: bool
            If False then print a log, if True then return a pd.DataFrame

        Returns
        -------
        None or pd.DataFrame.
        """
        n_errors = self.document.getNumErrors()
        if n_errors == 0:
            return None

        error_log = list()
        for i in range(n_errors):
            e = self.document.getError(i)

            error_entry = {
                SMBL_ERROR_NUMBER: i,
                SMBL_ERROR_CATEGORY: e.getCategoryAsString(),
                SMBL_ERROR_SEVERITY: e.getSeverity(),
                SMBL_ERROR_DESCRIPTION: e.getShortMessage(),
                SMBL_ERROR_MESSAGE: e.getMessage(),
            }

            error_log.append(error_entry)
        error_log = pd.DataFrame(error_log)

        if reduced_log:
            error_log = (
                error_log[
                    [SMBL_ERROR_CATEGORY, SMBL_ERROR_SEVERITY, SMBL_ERROR_MESSAGE]
                ]
                .groupby([SMBL_ERROR_CATEGORY, SMBL_ERROR_SEVERITY])
                .count()
            )

        if return_df:
            return error_log
        else:
            if reduced_log:
                headers = [f"{SMBL_ERROR_CATEGORY}, {SMBL_ERROR_SEVERITY}", "count"]
            else:
                headers = [
                    SMBL_ERROR_CATEGORY,
                    SMBL_ERROR_SEVERITY,
                    SMBL_ERROR_DESCRIPTION,
                ]
                error_log = error_log[headers]

            utils.style_df(error_log, headers=headers)

            return None


class SBML_reaction:
    """
    System Biology Markup Language Model Reactions.

    Attributes
    ----------
    reaction_dict: dict
        dictionary of reaction-level attributes, id, name, identifiers
    species: pd.DataFrame
        table of substrates, products, and modifiers

    """

    def __init__(
        self,
        sbml_reaction: libsbml.Reaction,
    ) -> None:
        """
        Convenience class for working with sbml reactions
        """
        reaction_dict = {
            SMBL_REACTION_DICT_ID: sbml_reaction.getId(),
            SMBL_REACTION_DICT_NAME: sbml_reaction.getName(),
            SMBL_REACTION_DICT_IDENTIFIERS: identifiers.cv_to_Identifiers(
                sbml_reaction
            ),
            SMBL_REACTION_DICT_SOURCE: source.Source(init=True),
            SMBL_REACTION_DICT_IS_REVERSIBLE: sbml_reaction.getReversible(),
        }

        self.reaction_dict = reaction_dict

        # process reaction species
        reaction_species = list()
        # save modifiers
        for i in range(sbml_reaction.getNumModifiers()):
            spec = sbml_reaction.getModifier(i)
            spec_dict = {
                SMBL_REACTION_SPEC_RSC_ID: spec.getId(),
                SMBL_REACTION_SPEC_SC_ID: spec.getSpecies(),
                SMBL_REACTION_SPEC_STOICHIOMETRY: 0,
                SMBL_REACTION_SPEC_SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)

        rxn_fbc = sbml_reaction.getPlugin("fbc")
        # check for gene products associated with the FBC L3 extension
        if rxn_fbc is not None:
            gene_products = list()
            gpa = rxn_fbc.getGeneProductAssociation()
            if gpa is not None:
                gpaa = gpa.getAssociation()
                if hasattr(gpaa, SBML_REACTION_ATTR_GET_GENE_PRODUCT):
                    gene_products.append(_get_gene_product_dict(gpaa))
                else:
                    for i in range(gpaa.getNumAssociations()):
                        gpaaa = gpaa.getAssociation(i)
                        if hasattr(gpaaa, SBML_REACTION_ATTR_GET_GENE_PRODUCT):
                            gene_products.append(_get_gene_product_dict(gpaaa))
                        else:
                            for i in range(gpaaa.getNumAssociations()):
                                gpaaaa = gpaaa.getAssociation(i)
                                if hasattr(gpaaaa, SBML_REACTION_ATTR_GET_GENE_PRODUCT):
                                    gene_products.append(_get_gene_product_dict(gpaaaa))
                                else:
                                    for i in range(gpaa.getNumAssociations()):
                                        gpaaaaa = gpaaaa.getAssociation(i)
                                        if hasattr(
                                            gpaaaaa, SBML_REACTION_ATTR_GET_GENE_PRODUCT
                                        ):
                                            gene_products.append(
                                                _get_gene_product_dict(gpaaaaa)
                                            )
                                        else:
                                            logger.warning(
                                                "gene annotations nested deeper than 4 levels, ignoring"
                                            )
                                            continue
            # de-duplicate
            gene_products = list(
                {d[SMBL_REACTION_SPEC_SC_ID]: d for d in gene_products}.values()
            )
            reaction_species = reaction_species + gene_products

        # save reactants
        for i in range(sbml_reaction.getNumReactants()):
            spec = sbml_reaction.getReactant(i)
            spec_dict = {
                SMBL_REACTION_SPEC_RSC_ID: spec.getId(),
                SMBL_REACTION_SPEC_SC_ID: spec.getSpecies(),
                SMBL_REACTION_SPEC_STOICHIOMETRY: -1 * spec.getStoichiometry(),
                SMBL_REACTION_SPEC_SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)
        # save products
        for i in range(sbml_reaction.getNumProducts()):
            spec = sbml_reaction.getProduct(i)
            spec_dict = {
                SMBL_REACTION_SPEC_RSC_ID: spec.getId(),
                SMBL_REACTION_SPEC_SC_ID: spec.getSpecies(),
                SMBL_REACTION_SPEC_STOICHIOMETRY: spec.getStoichiometry(),
                SMBL_REACTION_SPEC_SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)

        self.species = pd.DataFrame(reaction_species).set_index(
            SMBL_REACTION_SPEC_RSC_ID
        )


def sbml_df_from_sbml(self, sbml_model: SBML):
    # specify compartments

    compartments = list()
    for i in range(sbml_model.model.getNumCompartments()):
        comp = sbml_model.model.getCompartment(i)

        if not comp.getCVTerms():
            logger.warning(
                f"Compartment {comp.getId()} has empty CVterms, mapping its c_Identifiers from the Compartment dict"
            )

            comp_name = comp.getName()
            mapped_compartment_key = [
                compkey
                for compkey, mappednames in constants.COMPARTMENT_ALIASES.items()
                if comp_name in mappednames
            ]

            if len(mapped_compartment_key) == 0:
                logger.warning(
                    f"No GO compartment for {comp_name} is mapped, use the generic cellular_component's GO id"
                )
                compartments.append(
                    {
                        SBML_COMPARTMENT_DICT_ID: comp.getId(),
                        SBML_COMPARTMENT_DICT_NAME: comp.getName(),
                        SBML_COMPARTMENT_DICT_IDENTIFIERS: identifiers.Identifiers(
                            [
                                identifiers.format_uri(
                                    uri=identifiers.create_uri_url(
                                        ontology=constants.ONTOLOGIES.GO,
                                        identifier=constants.COMPARTMENTS_GO_TERMS[
                                            "CELLULAR_COMPONENT"
                                        ],
                                    ),
                                    biological_qualifier_type=BQB.BQB_IS,
                                )
                            ]
                        ),
                        SBML_COMPARTMENT_DICT_SOURCE: source.Source(init=True),
                    }
                )

            if len(mapped_compartment_key) > 0:
                if len(mapped_compartment_key) > 1:
                    logger.warning(
                        f"More than one GO compartments for {comp_name} are mapped, using the first one"
                    )
                compartments.append(
                    {
                        SBML_COMPARTMENT_DICT_ID: comp.getId(),
                        SBML_COMPARTMENT_DICT_NAME: comp.getName(),
                        SBML_COMPARTMENT_DICT_IDENTIFIERS: identifiers.Identifiers(
                            [
                                identifiers.format_uri(
                                    uri=identifiers.create_uri_url(
                                        ontology=constants.ONTOLOGIES.GO,
                                        identifier=constants.COMPARTMENTS_GO_TERMS[
                                            mapped_compartment_key[0]
                                        ],
                                    ),
                                    biological_qualifier_type=BQB.IS,
                                )
                            ]
                        ),
                        SBML_COMPARTMENT_DICT_SOURCE: source.Source(init=True),
                    }
                )

        else:
            compartments.append(
                {
                    SBML_COMPARTMENT_DICT_ID: comp.getId(),
                    SBML_COMPARTMENT_DICT_NAME: comp.getName(),
                    SBML_COMPARTMENT_DICT_IDENTIFIERS: identifiers.cv_to_Identifiers(
                        comp
                    ),
                    SBML_COMPARTMENT_DICT_SOURCE: source.Source(init=True),
                }
            )

    self.compartments = pd.DataFrame(compartments).set_index(SBML_COMPARTMENT_DICT_ID)

    # create a species df
    comp_species_df = setup_cspecies(sbml_model)

    # find unique species and create a table
    consensus_species_df = comp_species_df.copy()
    consensus_species_df.index.names = [SBML_SPECIES_DICT_ID]
    consensus_species, species_lookup = consensus.reduce_to_consensus_ids(
        consensus_species_df,
        {"pk": SBML_SPECIES_DICT_ID, "id": SBML_SPECIES_DICT_IDENTIFIERS},
    )

    # create a table of unique molecular species
    consensus_species.index.name = SBML_SPECIES_DICT_ID
    consensus_species[SBML_SPECIES_DICT_NAME] = [
        re.sub("\\[.+\\]", "", x).strip()
        for x in consensus_species[SBML_COMPARTMENTALIZED_SPECIES_DICT_NAME]
    ]
    consensus_species = consensus_species.drop(
        [SBML_COMPARTMENTALIZED_SPECIES_DICT_NAME, SBML_COMPARTMENT_DICT_ID], axis=1
    )
    consensus_species["s_Source"] = [
        source.Source(init=True) for x in range(0, consensus_species.shape[0])
    ]

    self.species = consensus_species[self.schema["species"]["vars"]]

    self.compartmentalized_species = comp_species_df.join(species_lookup).rename(
        columns={"new_id": SBML_SPECIES_DICT_ID}
    )[self.schema["compartmentalized_species"]["vars"]]

    # specify reactions

    reactions = list()
    reaction_species = list()
    for i in range(sbml_model.model.getNumReactions()):
        rxn = SBML_reaction(sbml_model.model.getReaction(i))
        reactions.append(rxn.reaction_dict)

        rxn_specs = rxn.species
        rxn_specs[SMBL_REACTION_DICT_ID] = rxn.reaction_dict[SMBL_REACTION_DICT_ID]
        reaction_species.append(rxn_specs)

    self.reactions = pd.DataFrame(reactions).set_index(SMBL_REACTION_DICT_ID)

    reaction_species_df = pd.concat(reaction_species)
    # add an index if reaction species didn't have IDs in the .sbml
    if all([v == "" for v in reaction_species_df.index.tolist()]):
        reaction_species_df = (
            reaction_species_df.reset_index(drop=True)
            .assign(
                rsc_id=sbml_dfs_utils.id_formatter(
                    range(reaction_species_df.shape[0]), SMBL_REACTION_SPEC_RSC_ID
                )
            )
            .set_index(SMBL_REACTION_SPEC_RSC_ID)
        )

    self.reaction_species = reaction_species_df

    return self


def setup_cspecies(sbml_model: SBML) -> pd.DataFrame:
    """
    Setup Compartmentalized Species

    Read all compartmentalized species from a model
    and setup as a pd.DataFrame.
    This operation is functionalized to test the subsequent call of
    consensus.reduce_to_consensus_ids()
    which collapses compartmentalized_species -> species
    based on shared identifiers.
    """
    comp_species = list()
    for i in range(sbml_model.model.getNumSpecies()):
        spec = sbml_model.model.getSpecies(i)

        spec_dict = {
            SMBL_REACTION_SPEC_SC_ID: spec.getId(),
            SBML_COMPARTMENTALIZED_SPECIES_DICT_NAME: spec.getName(),
            SBML_COMPARTMENT_DICT_ID: spec.getCompartment(),
            SBML_SPECIES_DICT_IDENTIFIERS: identifiers.cv_to_Identifiers(spec),
            SBML_COMPARTMENTALIZED_SPECIES_DICT_SOURCE: source.Source(init=True),
        }

        comp_species.append(spec_dict)

    mplugin = sbml_model.model.getPlugin("fbc")

    # add geneproducts defined using L3 FBC extension
    if mplugin is not None:
        for i in range(mplugin.getNumGeneProducts()):
            gene_product = mplugin.getGeneProduct(i)

            gene_dict = {
                SMBL_REACTION_SPEC_SC_ID: gene_product.getId(),
                SBML_COMPARTMENTALIZED_SPECIES_DICT_NAME: (
                    gene_product.getName()
                    if gene_product.isSetName()
                    else gene_product.getLabel()
                ),
                # use getLabel() to accomendate sbml model (e.g. HumanGEM.xml) with no fbc:name attribute
                # Recon3D.xml has both fbc:label and fbc:name attributes, with gene name in fbc:nam
                SBML_COMPARTMENT_DICT_ID: None,
                SBML_SPECIES_DICT_IDENTIFIERS: identifiers.cv_to_Identifiers(
                    gene_product
                ),
                SBML_COMPARTMENTALIZED_SPECIES_DICT_SOURCE: source.Source(init=True),
            }

            comp_species.append(gene_dict)

    return pd.DataFrame(comp_species).set_index(SMBL_REACTION_SPEC_SC_ID)


def add_sbml_annotations(
    sbml_model: SBML, annotations: pd.DataFrame, save_path: str
) -> None:
    """
    Add SBML Annotations

    Add additional identifiers to an sbml file and save the updated document

    Parameters:
    sbml_model: SBML
        A .sbml model
    annotations: pd.DataFrame
        A table of annotations to add containing an "id" matching the
        primary key of an entity, "type" matching the type of entity,
        and "uri" representing the annotation to add.
    save_path: str
        Path to save the model to

    Returns:
    None
    """

    logger.warning(
        "add_sbml_annotations is deprecated and may be removed in a future version of rcpr; "
        "we are now adding these annotation during ingestion by sbml.sbml_df_from_sbml() rather "
        "than directly appending them to the raw .sbml"
    )

    if not isinstance(sbml_model, SBML):
        raise TypeError("sbml_model must be an SBML object")

    if not isinstance(annotations, pd.DataFrame):
        raise TypeError("annotations must be a pd.DataFrame")

    for i in range(0, annotations.shape[0]):
        annot_type = annotations["type"][i]

        if annot_type == "species":
            entity_fxn = SBML_ANNOTATION_METHOD_GET_SPECIES
        elif annot_type == "compartment":
            entity_fxn = SBML_ANNOTATION_METHOD_GET_COMPARTMENT
        elif annot_type == "reaction":
            entity_fxn = SBML_ANNOTATION_METHOD_GET_REACTION
        else:
            raise ValueError(
                f"{annot_type} is not a valid annotation type,"
                " valid types are species, compartment, and reaction"
            )
        # access the node to modify
        entity_fxn_method = getattr(sbml_model.model, entity_fxn)
        entity_node = entity_fxn_method(annotations["id"][i])

        # TO DO - check for a valid entity_node in case id is not found

        # set meta-id if there isn't one; required to add a node
        if not entity_node.isSetMetaId():
            add_metaid_code = entity_node.setMetaId(annotations["id"][i])

            if add_metaid_code != libsbml.LIBSBML_OPERATION_SUCCESS:
                raise ValueError(
                    f"adding metaId to {annotations['id'][i]} failed"
                    f" with return code {add_metaid_code} "
                    f"({libsbml.OperationReturnValue_toString(add_metaid_code).strip()})"
                )

        # create a controlled vocabulary term
        cv = libsbml.CVTerm()
        cv.setQualifierType(libsbml.BIOLOGICAL_QUALIFIER)
        cv.setBiologicalQualifierType(libsbml.BQB_IS_VERSION_OF)

        add_resource_code = cv.addResource(annotations["uri"][i])
        if add_resource_code != libsbml.LIBSBML_OPERATION_SUCCESS:
            raise ValueError(
                "adding resource to CV term returned code"
                f" {add_resource_code} "
                f"({libsbml.OperationReturnValue_toString(add_resource_code).strip()})"
                f" rather than {libsbml.LIBSBML_OPERATION_SUCCESS} when "
                f"adding {annotations['uri'][i]} to {annotations['id'][i]}"
            )

        add_cv_code = entity_node.addCVTerm(cv)
        if add_cv_code != libsbml.LIBSBML_OPERATION_SUCCESS:
            raise ValueError(
                f"adding CV to entity returned code {add_cv_code} "
                f"({libsbml.OperationReturnValue_toString(add_cv_code).strip()})"
                f" rather than {libsbml.LIBSBML_OPERATION_SUCCESS} when adding"
                f" {annotations['uri'][i]} to {annotations['id'][i]}"
            )

    libsbml.writeSBML(sbml_model.document, save_path)

    return None


def _get_gene_product_dict(gp):
    """Read a gene product node from an sbml file."""
    return {
        SMBL_REACTION_SPEC_RSC_ID: gp.getId(),
        SMBL_REACTION_SPEC_SC_ID: gp.getGeneProduct(),
        SMBL_REACTION_SPEC_STOICHIOMETRY: 0,
        SMBL_REACTION_SPEC_SBO_TERM: gp.getSBOTermID(),
    }



from __future__ import annotations

import logging

import pandas as pd
from napistu import constants
from napistu import identifiers
from napistu import sbml_dfs_core
from napistu import source
from napistu import utils
from napistu.rpy2 import callr
from napistu.rpy2 import report_r_exceptions
from napistu.rpy2 import warn_if_no_rpy2

from napistu.constants import SBML_DFS
from napistu.constants import BQB
from napistu.constants import IDENTIFIERS
from napistu.constants import ONTOLOGIES
from napistu.constants import ONTOLOGY_ALIASES
from napistu.rpy2.constants import BIOC_VALID_EXPANDED_SPECIES_ONTOLOGIES
from napistu.rpy2.constants import BIOC_DOGMATIC_MAPPING_ONTOLOGIES
from napistu.rpy2.constants import BIOC_PROTEIN_ONTOLOGIES
from napistu.rpy2.constants import BIOC_NAME_ONTOLOGIES
from napistu.rpy2.constants import BIOC_GENE_ONTOLOGIES  # noqa
from napistu.rpy2.constants import BIOC_NOMENCLATURE

logger = logging.getLogger(__name__)


@warn_if_no_rpy2
@report_r_exceptions
def expand_identifiers(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    id_type: str,
    species: str,
    expanded_ontologies: list[str],
    r_paths: str | None = None,
) -> pd.Series:
    """
    Expand Identifiers

    Update a table's identifiers to include additional related ontologies

    Ontologies are pulled from the bioconductor "org" packages. This is effective, but inelegant.

    Parameters
    ----------
    sbml_dfs : SBML_dfs
        A relational pathway model built around reactions interconverting compartmentalized species.
    id_type: str
        Identifiers to expand: species, compartments, or reactions
    species: str
        Species name
    expanded_ontologies: list
        Ontologies to add or complete
    r_paths: str
        Path to an R packages directory

    Returns
    -------
    a pd.Series with identifiers as the index and updated Identifiers objects as values
    """

    if not isinstance(sbml_dfs, sbml_dfs_core.SBML_dfs):
        raise TypeError("sbml_dfs is not an sbml_dfs_core.SBML_dfs object")

    # pull out all identifiers as a pd.DataFrame
    all_entity_identifiers = sbml_dfs.get_identifiers(id_type)
    if not isinstance(all_entity_identifiers, pd.DataFrame):
        raise TypeError("all_entity_identifiers must be a pandas DataFrame")

    if id_type == "species":
        all_entity_identifiers = _check_species_identifiers_entrez_gene_ontology(
            all_entity_identifiers
        )

        valid_expanded_ontologies = BIOC_VALID_EXPANDED_SPECIES_ONTOLOGIES
    elif id_type in ["reactions", "compartments"]:
        raise NotImplementedError(
            f"No converters implemented to expand {id_type} annotations"
        )
    else:
        raise ValueError(f"{id_type} is an invalid id_type")

    invalid_expanded_ontologies = set(expanded_ontologies).difference(
        valid_expanded_ontologies
    )

    if len(invalid_expanded_ontologies) != 0:
        raise NotImplementedError(
            f"No converters implemented to expand {id_type} annotations to {', '.join(invalid_expanded_ontologies)}"
        )

    # find entries in valid_expanded_ontologies which are already present
    # these are the entries that will be used to expand to other ontologies
    # or fill in ontologies with incomplete annotations
    starting_ontologies = valid_expanded_ontologies.intersection(
        set(all_entity_identifiers["ontology"])
    )

    if len(starting_ontologies) == 0:
        raise ValueError(f"No ontologies with {id_type} converters are present")

    required_conversion_ontologies = set(starting_ontologies).union(
        set(expanded_ontologies)
    )

    # pull down entrez ids + mapping to other ontologies
    mapping_ontologies = required_conversion_ontologies.intersection(
        BIOC_VALID_EXPANDED_SPECIES_ONTOLOGIES
    )

    mappings_dict = create_bioconductor_mapping_tables(
        mappings=mapping_ontologies, species=species, r_paths=r_paths
    )

    # start with entrez IDs (since all other ontologies are mapped to them in the
    # bioconductor "org" packages)

    # get these values by just looking up the mappings between entrez genes and genomic loci
    running_ids = merge_bioconductor_mappings(mappings_dict, mapping_ontologies)

    # map from existing ontologies to expanded ontologies
    ontology_mappings = list()
    # starting w/
    for start in starting_ontologies:
        # ending w/
        for end in expanded_ontologies:
            if start == end:
                continue
            lookup = (
                running_ids[[start, end]]
                .rename(columns={start: IDENTIFIERS.IDENTIFIER, end: "new_identifier"})
                .assign(ontology=start)
                .assign(new_ontology=end)
            )
            ontology_mappings.append(lookup)

    ontology_mappings_df = pd.concat(ontology_mappings).dropna()

    # old identifiers joined with new identifiers

    # first, define the names of keys and ids
    table_pk_var = sbml_dfs.schema[id_type]["pk"]
    table_id_var = sbml_dfs.schema[id_type]["id"]

    # retain bqb terms to define how an identifier is related to sid
    # this relation will be preserved for the new ids

    merged_identifiers = all_entity_identifiers[
        [
            table_pk_var,
            IDENTIFIERS.ONTOLOGY,
            IDENTIFIERS.IDENTIFIER,
            IDENTIFIERS.BQB,
        ]
    ].merge(ontology_mappings_df)

    # new, possibly redundant identifiers
    new_identifiers = merged_identifiers[
        [table_pk_var, "new_ontology", "new_identifier", IDENTIFIERS.BQB]
    ].rename(
        columns={
            "new_ontology": IDENTIFIERS.ONTOLOGY,
            "new_identifier": IDENTIFIERS.IDENTIFIER,
        }
    )

    expanded_identifiers_df = (
        pd.concat(
            [
                all_entity_identifiers[
                    [
                        table_pk_var,
                        IDENTIFIERS.ONTOLOGY,
                        IDENTIFIERS.IDENTIFIER,
                        IDENTIFIERS.URL,
                        IDENTIFIERS.BQB,
                    ]
                ],
                new_identifiers,
                # ignore new identifier if it already exists
            ]
        )
        # remove duplicated identifiers
        .groupby([table_pk_var, IDENTIFIERS.ONTOLOGY, IDENTIFIERS.IDENTIFIER])
        .first()
        .reset_index()
        .set_index(table_pk_var)
    )

    # create a dictionary of new Identifiers objects
    expanded_identifiers_dict = {
        i: _expand_identifiers_new_entries(i, expanded_identifiers_df)
        for i in expanded_identifiers_df.index.unique()
    }

    output = pd.Series(expanded_identifiers_dict).rename(table_id_var)
    output.index.name = table_pk_var

    return output


@warn_if_no_rpy2
@report_r_exceptions
def create_bioconductor_mapping_tables(
    mappings: set[str], species: str, r_paths: str | None = None
) -> dict[str, pd.DataFrame]:
    """
    Create Bioconductor Mapping Tables

    Creating a dictionary of mappings between entrez and other ontologies.

    Args:
        mappings (set):
            A set of ontologies to work with. The valid ontologies are:
            "ensembl_gene", "ensembl_transcript", and "uniprot".
        species (str):
            The organismal species that we are working with (e.g., Homo sapiens).
        r_paths (str, optional):
            Optional path to a library of R packages.

    Returns:
        mappings_dict (dict):
            A table of entrez ids, and tables mapping from each ontology in "mappings" to entrez.

    """

    if not isinstance(mappings, set):
        raise TypeError(f"mappings must be a set, but got {type(mappings).__name__}")
    if not isinstance(species, str):
        raise TypeError(f"species must be a str, but got {type(species).__name__}")

    logger.info(
        f"Creating mapping tables from entrez genes to/from {', '.join(mappings)}"
    )

    invalid_mappings = set(mappings).difference(BIOC_VALID_EXPANDED_SPECIES_ONTOLOGIES)

    if len(invalid_mappings) > 0:
        raise ValueError(
            f"{len(invalid_mappings)} mappings could not be created: {', '.join(invalid_mappings)}.\n"
            f"The valid mappings are {', '.join(BIOC_VALID_EXPANDED_SPECIES_ONTOLOGIES)}"
        )

    mappings_dict = dict()

    # all mappings are with respect to entrez. so we will always want to obtain entrez ids
    mappings_dict[ONTOLOGIES.NCBI_ENTREZ_GENE] = (
        callr.r_dataframe_to_pandas(
            callr.bioconductor_org_r_function(
                BIOC_NOMENCLATURE.CHR_TBL, species, r_paths=None
            )
        )
        .drop(BIOC_NOMENCLATURE.CHROMOSOME, axis=1)
        .rename(
            columns={BIOC_NOMENCLATURE.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE}
        )
        .set_index(ONTOLOGIES.NCBI_ENTREZ_GENE)
    )

    if ONTOLOGIES.ENSEMBL_GENE in mappings:
        # "entrez <> ensembl genes"
        mappings_dict[ONTOLOGIES.ENSEMBL_GENE] = (
            callr.r_dataframe_to_pandas(
                callr.bioconductor_org_r_function(
                    BIOC_NOMENCLATURE.ENSG_TBL, species, r_paths=r_paths
                )
            )
            .rename(
                columns={
                    BIOC_NOMENCLATURE.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE,
                    BIOC_NOMENCLATURE.ENSEMBL_GENE: ONTOLOGIES.ENSEMBL_GENE,
                }
            )
            .set_index(ONTOLOGIES.NCBI_ENTREZ_GENE)
        )

    if ONTOLOGIES.ENSEMBL_TRANSCRIPT in mappings:
        # "entrez <> ensembl transcripts"
        mappings_dict[ONTOLOGIES.ENSEMBL_TRANSCRIPT] = (
            callr.r_dataframe_to_pandas(
                callr.bioconductor_org_r_function(
                    BIOC_NOMENCLATURE.ENST_TBL, species, r_paths=r_paths
                )
            )
            .rename(
                columns={
                    BIOC_NOMENCLATURE.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE,
                    BIOC_NOMENCLATURE.ENSEMBL_TRANSCRIPT: ONTOLOGIES.ENSEMBL_TRANSCRIPT,
                }
            )
            .set_index(ONTOLOGIES.NCBI_ENTREZ_GENE)
        )

    if ONTOLOGIES.ENSEMBL_PROTEIN in mappings:
        # "entrez <> ensembl proteins"
        mappings_dict[ONTOLOGIES.ENSEMBL_PROTEIN] = (
            callr.r_dataframe_to_pandas(
                callr.bioconductor_org_r_function(
                    BIOC_NOMENCLATURE.ENSP_TBL, species, r_paths=r_paths
                )
            )
            .rename(
                columns={
                    BIOC_NOMENCLATURE.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE,
                    BIOC_NOMENCLATURE.ENSEMBL_PROTEIN: ONTOLOGIES.ENSEMBL_PROTEIN,
                }
            )
            .set_index(ONTOLOGIES.NCBI_ENTREZ_GENE)
        )

    if ONTOLOGIES.UNIPROT in mappings:
        # "entrez <> uniprot"
        mappings_dict[ONTOLOGIES.UNIPROT] = (
            callr.r_dataframe_to_pandas(
                callr.bioconductor_org_r_function(
                    BIOC_NOMENCLATURE.UNIPROT_TBL, species, r_paths=r_paths
                )
            )
            .rename(
                columns={
                    BIOC_NOMENCLATURE.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE,
                    BIOC_NOMENCLATURE.UNIPROT: ONTOLOGIES.UNIPROT,
                }
            )
            .set_index(ONTOLOGIES.NCBI_ENTREZ_GENE)
        )

    if ONTOLOGIES.GENE_NAME in mappings:
        # "entrez <> gene name"
        mappings_dict[ONTOLOGIES.GENE_NAME] = (
            callr.r_dataframe_to_pandas(
                callr.bioconductor_org_r_function(
                    BIOC_NOMENCLATURE.NAME_TBL, species, r_paths=r_paths
                )
            )
            .rename(
                columns={
                    BIOC_NOMENCLATURE.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE,
                    BIOC_NOMENCLATURE.GENE_NAME: ONTOLOGIES.GENE_NAME,
                }
            )
            .set_index(ONTOLOGIES.NCBI_ENTREZ_GENE)
        )

    if ONTOLOGIES.SYMBOL in mappings:
        # "entrez <> gene symbol"
        mappings_dict[ONTOLOGIES.SYMBOL] = (
            callr.r_dataframe_to_pandas(
                callr.bioconductor_org_r_function(
                    BIOC_NOMENCLATURE.SYMBOL_TBL, species, r_paths=r_paths
                )
            )
            .rename(
                columns={
                    BIOC_NOMENCLATURE.NCBI_ENTREZ_GENE: ONTOLOGIES.NCBI_ENTREZ_GENE,
                    BIOC_NOMENCLATURE.SYMBOL: ONTOLOGIES.SYMBOL,
                }
            )
            .set_index(ONTOLOGIES.NCBI_ENTREZ_GENE)
        )

    return mappings_dict


def merge_bioconductor_mappings(
    mappings_dict: dict, mapping_ontologies: set[str]
) -> pd.DataFrame:
    """Combine multiple ontologies by recursively joining on Entrez Gene"""

    running_ids = mappings_dict[ONTOLOGIES.NCBI_ENTREZ_GENE]

    for mapping in mapping_ontologies:
        logger.debug(f"adding entries for {mapping} to running_ids")
        mapping_df = mappings_dict[mapping]

        running_ids = running_ids.join(mapping_df)

    running_ids = running_ids.reset_index()

    return running_ids


def stack_bioconductor_mappings(
    mappings_dict: dict[str, pd.DataFrame], mapping_ontologies: set[str]
) -> pd.DataFrame:
    """
    Stack Bioconductor Mappings

    Convert a dict of mappings between entrez identifiers and other identifiers to a single table.

    Args:
        mappings_dict (dict):
            A dictionary containing mappings between entrez and other ontologies.
        mapping_ontologies (set):
            A set of mappings to combine.

    Returns:
        mappings_df (pd.DataFrame):
            A table containing entrez_gene_id, ontology, and identifier.
    """

    mappings_list = list()
    for ont in mapping_ontologies:
        one_mapping_df = (
            mappings_dict[ont].assign(ontology=ont).rename({ont: "identifier"}, axis=1)
        )

        mappings_list.append(one_mapping_df)

    return pd.concat(mappings_list)


def _check_species_identifiers_entrez_gene_ontology(
    entity_identifiers_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Check whether species ontologies contain ncbigene or ncbi_gene
    If so, replaced them to ncbi_entrez_gene.
    Return: entity_identifiers_df with proper gene ontology types.
    """

    intersect_gene_onto = set(entity_identifiers_df["ontology"]).intersection(
        ONTOLOGY_ALIASES.NCBI_ENTREZ_GENE
    )

    # if entity_identifiers_df contains members of ENTREZ_ONTOLOGY_ALIASES,
    # replace to ncbi_entrez_gene
    if intersect_gene_onto:
        logger.info(
            f" Replace unmatching ontology {', '.join(intersect_gene_onto)} to {ONTOLOGIES.NCBI_ENTREZ_GENE}."
        )

        filtered_onto_df = entity_identifiers_df[
            entity_identifiers_df["ontology"].isin(list(intersect_gene_onto))
        ]

        entity_identifiers_df.loc[filtered_onto_df.index, "ontology"] = (
            ONTOLOGIES.NCBI_ENTREZ_GENE
        )

    return entity_identifiers_df


def update_expanded_identifiers(
    model: sbml_dfs_core.SBML_dfs, id_type: str, expanded_ids: pd.Series
) -> sbml_dfs_core.SBML_dfs:
    """Update the expanded identifiers for a model.

    Args:
        model (sbml_dfs_core.SBML_dfs): _description_
        id_type (str): _description_
        expanded_ids (str): _description_
    """
    ids = getattr(model, id_type)

    # make sure expanded_ids and original model.species have same number of s_ids
    # if a s_id only in model.species, adding it to expanded_ids.
    if ids.shape[0] != expanded_ids.shape[0]:
        matched_expanded_ids = expanded_ids.combine_first(ids[SBML_DFS.S_IDENTIFIERS])
        logger.debug(
            f"{ids.shape[0] - expanded_ids.shape[0]} "
            "ids are not included in expanded ids"
        )
    else:
        matched_expanded_ids = expanded_ids

    updated_ids = ids.drop(SBML_DFS.S_IDENTIFIERS, axis=1).join(
        pd.DataFrame(matched_expanded_ids)
    )

    setattr(model, id_type, updated_ids)

    return model


def create_dogmatic_sbml_dfs(
    species: str, r_paths: str | None = None
) -> sbml_dfs_core.SBML_dfs:
    """
    Create Dogmatic SMBL_DFs

    Create an SBML_dfs model which is pretty much just proteins and no
    reactions, as well as annotations linking proteins to genes, and
    creating nice labels for genes/proteins.

    Args:
        species (str):
            An organismal species (e.g., Homo sapiens)
        r_paths (str or None)
            Optional, p]ath to an R packages directory

    Returns:
        dogmatic_sbml_dfs (sbml.SBML_dfs)
            A pathway model which (pretty much) just contains proteins and
            diverse identifiers
    """

    dogmatic_mappings = connect_dogmatic_mappings(species)

    logger.info("Creating inputs for sbml_dfs_from_edgelist()")

    # format entries for sbml_dfs_from_edgelist()
    species_df = dogmatic_mappings["cluster_consensus_identifiers_df"].join(
        dogmatic_mappings["s_name_series"]
    )

    # stub required but invariant variables
    compartments_df = sbml_dfs_core._stub_compartments()
    interaction_source = source.Source(init=True)

    # interactions table. This is required to create the sbml_dfs but we'll drop the info later
    interaction_edgelist = species_df.rename(
        columns={
            "s_name": "upstream_name",
            SBML_DFS.S_IDENTIFIERS: SBML_DFS.R_IDENTIFIERS,
        }
    )
    interaction_edgelist["downstream_name"] = interaction_edgelist["upstream_name"]
    interaction_edgelist["upstream_compartment"] = "cellular_component"
    interaction_edgelist["downstream_compartment"] = "cellular_component"
    interaction_edgelist["r_name"] = interaction_edgelist["upstream_name"]
    interaction_edgelist["sbo_term"] = constants.MINI_SBO_FROM_NAME["reactant"]
    interaction_edgelist["r_isreversible"] = False

    dogmatic_sbml_dfs = sbml_dfs_core.sbml_dfs_from_edgelist(
        interaction_edgelist=interaction_edgelist,
        species_df=species_df,
        compartments_df=compartments_df,
        interaction_source=interaction_source,
        upstream_stoichiometry=-1,
        downstream_stoichiometry=1,
        downstream_sbo_name="product",
    )

    # remove all reactions except 1 (so it still passes sbml_dfs.validate())
    # this self reaction will be removed when creating the graph
    dogmatic_sbml_dfs.remove_reactions(dogmatic_sbml_dfs.reactions.index.tolist()[1::])

    return dogmatic_sbml_dfs


def connect_dogmatic_mappings(species: str, r_paths: str | None = None) -> dict:
    """
    Connect Dogmatic Mappings

    Merge all ontologies into greedy clusters based on shared associations to entrez ids

    Args:
        species (str):
            An organismal species (e.g., Homo sapiens)
        r_paths (str or None)
            Optional, p]ath to an R packages directory

    Returns:
        dict with:
        - s_name_series: a series where the index is distinct molecular species and the values are names.
        - cluster_consensus_identifiers_df: a pd.DataFrame where the index is distinct molecular species
          and values are identifiers objects.
    """

    mappings_dict = create_bioconductor_mapping_tables(
        mappings=BIOC_DOGMATIC_MAPPING_ONTOLOGIES,
        species=species,
        r_paths=r_paths,
    )

    protein_mappings = stack_bioconductor_mappings(
        mappings_dict, set(BIOC_PROTEIN_ONTOLOGIES)
    )

    # apply greedy graph-based clustering to connect proteins with a common mapping to entrez
    edgelist_df = utils.format_identifiers_as_edgelist(
        protein_mappings, [IDENTIFIERS.ONTOLOGY, IDENTIFIERS.IDENTIFIER]
    )
    connected_indices = utils.find_weakly_connected_subgraphs(
        edgelist_df[["ind", "id"]]
    )

    # add clusters to proteins. Each cluster will be a distinct molecular species
    protein_mappings_w_clusters = protein_mappings.reset_index().merge(
        connected_indices
    )

    # combine entrez + cluster so we can pass cluster to non-protein attributes
    entrez_clusters = protein_mappings_w_clusters[
        [ONTOLOGIES.NCBI_ENTREZ_GENE, "cluster"]
    ].drop_duplicates()
    other_ontologies = BIOC_DOGMATIC_MAPPING_ONTOLOGIES.difference(
        set(BIOC_PROTEIN_ONTOLOGIES)
    )
    other_mappings = stack_bioconductor_mappings(mappings_dict, other_ontologies)
    other_mappings_w_clusters = entrez_clusters.merge(
        other_mappings, left_on=ONTOLOGIES.NCBI_ENTREZ_GENE, right_index=True
    )

    possible_names = pd.concat(
        [
            protein_mappings_w_clusters.query(
                "ontology in @BIOC_NAME_ONTOLOGIES.keys()"
            ),
            other_mappings_w_clusters.query("ontology in @BIOC_NAME_ONTOLOGIES.keys()"),
        ]
    )[["cluster", IDENTIFIERS.ONTOLOGY, IDENTIFIERS.IDENTIFIER]]

    possible_names.loc[:, "ontology_preference"] = possible_names[
        IDENTIFIERS.ONTOLOGY
    ].map(BIOC_NAME_ONTOLOGIES)

    # remove possible names which are present in multiple clusters.
    # all clusters will need unique names to use sbml_dfs_from_edgelist()
    id_counts = (
        possible_names[["cluster", IDENTIFIERS.IDENTIFIER]]
        .drop_duplicates()
        .value_counts(IDENTIFIERS.IDENTIFIER)
    )
    possible_names = possible_names[
        ~possible_names[IDENTIFIERS.IDENTIFIER].isin(
            id_counts[id_counts > 1].index.tolist()
        )
    ]

    s_name_series = (
        utils._add_nameness_score(possible_names, IDENTIFIERS.IDENTIFIER)
        .sort_values(["ontology_preference", "nameness_score"])
        .groupby("cluster")
        .first()
        .rename(columns={IDENTIFIERS.IDENTIFIER: SBML_DFS.S_NAME})[SBML_DFS.S_NAME]
    )

    protein_ids = protein_mappings_w_clusters.assign(bqb=BQB.IS)[
        ["cluster", IDENTIFIERS.IDENTIFIER, IDENTIFIERS.ONTOLOGY, IDENTIFIERS.BQB]
    ]
    gene_ids = other_mappings_w_clusters.query(
        "ontology in @BIOC_GENE_ONTOLOGIES"
    ).assign(bqb=BQB.IS_ENCODED_BY)[
        ["cluster", IDENTIFIERS.IDENTIFIER, IDENTIFIERS.ONTOLOGY, IDENTIFIERS.BQB]
    ]
    entrez_ids = entrez_clusters.assign(
        ontology=ONTOLOGIES.NCBI_ENTREZ_GENE, bqb=BQB.IS_ENCODED_BY
    ).rename(columns={ONTOLOGIES.NCBI_ENTREZ_GENE: IDENTIFIERS.IDENTIFIER})[
        ["cluster", IDENTIFIERS.IDENTIFIER, IDENTIFIERS.ONTOLOGY, IDENTIFIERS.BQB]
    ]

    # combine all ids to setup a single cluster-level Identifiers
    all_ids = pd.concat([protein_ids, gene_ids, entrez_ids])
    all_ids.loc[:, IDENTIFIERS.URL] = [
        identifiers.create_uri_url(x, y)
        for x, y in zip(all_ids[IDENTIFIERS.ONTOLOGY], all_ids[IDENTIFIERS.IDENTIFIER])
    ]

    # create one Identifiers object for each new species
    cluster_consensus_identifiers = {
        k: identifiers.Identifiers(
            list(
                v[
                    [
                        IDENTIFIERS.ONTOLOGY,
                        IDENTIFIERS.IDENTIFIER,
                        IDENTIFIERS.URL,
                        IDENTIFIERS.BQB,
                    ]
                ]
                .reset_index(drop=True)
                .T.to_dict()
                .values()
            )
        )
        for k, v in all_ids.groupby("cluster")
    }

    cluster_consensus_identifiers_df = pd.DataFrame(
        cluster_consensus_identifiers, index=[SBML_DFS.S_IDENTIFIERS]
    ).T
    cluster_consensus_identifiers_df.index.name = "cluster"

    out_dict = {
        "s_name_series": s_name_series,
        "cluster_consensus_identifiers_df": cluster_consensus_identifiers_df,
    }

    return out_dict


@warn_if_no_rpy2
def _expand_identifiers_new_entries(
    sysid: str, expanded_identifiers_df: pd.DataFrame
) -> identifiers.Identifiers:
    """Expand Identifiers to include Bioconductor annotations"""
    entry = expanded_identifiers_df.loc[sysid]

    if type(entry) is pd.Series:
        sysis_id_list = [entry.to_dict()]
    else:
        # multiple annotations
        sysis_id_list = list(entry.reset_index(drop=True).T.to_dict().values())

    return identifiers.Identifiers(sysis_id_list)



"""
MCP (Model Context Protocol) Server for Napistu.
"""

import asyncio
import logging
import click
import click_logging

import napistu
from napistu.mcp.profiles import get_profile, ServerProfile
from napistu.mcp.server import create_server

logger = logging.getLogger(napistu.__name__)
click_logging.basic_config(logger)


@click.group()
def cli():
    """The Napistu MCP (Model Context Protocol) Server CLI"""
    pass


@click.group()
def server():
    """Start and manage MCP servers."""
    pass


@server.command(name="start")
@click.option(
    "--profile",
    type=click.Choice(["local", "remote", "full"]),
    default="remote",
    help="Predefined configuration profile",
)
@click.option("--server-name", type=str, help="Name of the MCP server")
@click_logging.simple_verbosity_option(logger)
def start_server(profile, server_name):
    """Start an MCP server with the specified profile."""
    # Collect configuration
    config = {}
    if server_name:
        config["server_name"] = server_name

    # Get profile with overrides
    server_profile = get_profile(profile, **config)

    # Create and start the server
    logger.info(f"Starting Napistu MCP Server with {profile} profile...")
    server = create_server(server_profile)
    asyncio.run(server.start())


@server.command(name="local")
@click.option(
    "--server-name", type=str, default="napistu-local", help="Name of the MCP server"
)
@click_logging.simple_verbosity_option(logger)
def start_local(server_name):
    """Start a local MCP server optimized for function execution."""
    # Get profile with overrides
    server_profile = get_profile("local", server_name=server_name)

    # Create and start the server
    logger.info("Starting Napistu local MCP Server...")
    server = create_server(server_profile)
    asyncio.run(server.start())


@server.command(name="remote")
@click.option(
    "--server-name", type=str, default="napistu-docs", help="Name of the MCP server"
)
@click.option("--codebase-path", type=str, help="Path to the Napistu codebase")
@click.option(
    "--docs-paths",
    type=str,
    help="Comma-separated list of paths to documentation files",
)
@click.option("--tutorials-path", type=str, help="Path to the tutorials directory")
@click_logging.simple_verbosity_option(logger)
def start_remote(server_name, tutorials_path):
    """Start a remote MCP server for documentation and codebase exploration."""
    # Collect configuration
    config = {"server_name": server_name}
    if tutorials_path:
        config["tutorials_path"] = tutorials_path

    # Get profile with overrides
    server_profile = get_profile("remote", **config)

    # Create and start the server
    logger.info("Starting Napistu remote MCP Server...")
    server = create_server(server_profile)
    asyncio.run(server.start())


@click.group()
def component():
    """Enable or disable specific MCP server components."""
    pass


@component.command(name="list")
def list_components():
    """List available MCP server components."""
    click.echo("Available MCP server components:")
    click.echo("  - documentation: Documentation components")
    click.echo("  - codebase: Codebase exploration components")
    click.echo("  - execution: Function execution components")
    click.echo("  - tutorials: Tutorial components")


@component.command(name="custom")
@click.option(
    "--enable-documentation/--disable-documentation",
    default=None,
    help="Enable/disable documentation components",
)
@click.option(
    "--enable-codebase/--disable-codebase",
    default=None,
    help="Enable/disable codebase exploration components",
)
@click.option(
    "--enable-execution/--disable-execution",
    default=None,
    help="Enable/disable function execution components",
)
@click.option(
    "--enable-tutorials/--disable-tutorials",
    default=None,
    help="Enable/disable tutorial components",
)
@click.option(
    "--server-name", type=str, default="napistu-custom", help="Name of the MCP server"
)
@click.option("--codebase-path", type=str, help="Path to the Napistu codebase")
@click.option(
    "--docs-paths",
    type=str,
    help="Comma-separated list of paths to documentation files",
)
@click.option("--tutorials-path", type=str, help="Path to the tutorials directory")
@click_logging.simple_verbosity_option(logger)
def custom_server(
    enable_documentation,
    enable_codebase,
    enable_execution,
    enable_tutorials,
    server_name,
):
    """Start an MCP server with custom component configuration."""
    # Collect configuration
    config = {"server_name": server_name}
    if enable_documentation is not None:
        config["enable_documentation"] = enable_documentation
    if enable_codebase is not None:
        config["enable_codebase"] = enable_codebase
    if enable_execution is not None:
        config["enable_execution"] = enable_execution
    if enable_tutorials is not None:
        config["enable_tutorials"] = enable_tutorials

    # Create a custom profile
    server_profile = ServerProfile(**config)

    # Create and start the server
    logger.info("Starting Napistu custom MCP Server...")
    server = create_server(server_profile)
    asyncio.run(server.start())


# Add command groups to the CLI
cli.add_command(server)
cli.add_command(component)

if __name__ == "__main__":
    cli()



"""Module containing functions to interoperate with rcpr's netcontextr functions"""

from __future__ import annotations

import logging
import os
from tempfile import NamedTemporaryFile
from typing import Any
from typing import Callable
from typing import Iterable

import pandas as pd
from napistu import sbml_dfs_core
from napistu import utils
from napistu.rpy2 import has_rpy2
from napistu.rpy2 import warn_if_no_rpy2

from napistu.rpy2.constants import COL_GENE
from napistu.rpy2.constants import COL_PROTEIN_1
from napistu.rpy2.constants import COL_PROTEIN_2
from napistu.rpy2.constants import FIELD_INTERACTIONS
from napistu.rpy2.constants import FIELD_GENES
from napistu.rpy2.constants import FIELD_REACTIONS
from napistu.rpy2.constants import COL_ROLE
from napistu.rpy2.constants import COL_REACTION_ID
from napistu.rpy2.constants import COL_STOICHIOMETRY
from napistu.rpy2.constants import NETCONTEXTR_ONTOLOGY
from napistu.rpy2.constants import NETCONTEXTR_SBO_MAP

if has_rpy2:
    from napistu.rpy2.callr import pandas_to_r_dataframe
    from rpy2.robjects import ListVector
    import rpy2.robjects as robjs

logger = logging.getLogger(__name__)


@warn_if_no_rpy2
def _none2null(none_obj):
    return robjs.r("NULL")


@warn_if_no_rpy2
def sbml_dfs_to_rcpr_string_graph(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    reaction_data: str = "string",
    identifier_ontology: str = "ensembl_gene",
    rescale_data: Callable[[pd.DataFrame], pd.DataFrame] | None = lambda x: x / 1000,
) -> ListVector:
    """Converts an sbml_dfs to a rcpr string graph

    This utility converts the sbml_dfs to the format returned
    by `rcpr::createStringGraph`.

    Args:
        sbml_dfs (SBML_dfs): the sbml_dfs from string.
            It is assumed that this sbml_dfs has only reactions with exactly
            two reactands and a 1:1 mapping between s_id and sc_id.
        reaction_data (str, optional): The reaction data that contains
            the string scores. Defaults to 'string'.
        identifier_ontology (str, optional): The ontology to use for the
            protein identifiers. Defaults to `ensembl_gene` (default in rcpr)
        rescale_data (Callable[pd.DataFrame], optional): A function to rescale
            the data. Defaults to lambda x: x/1000 (default in rcpr)

    Returns:
        This is a list of dataframes almost the same as `rcpr::createStringGraph`:
        - `genes`: a dataframe with column `gene`
            and the extra column `s_id`, `sc_id`
        - `interactions`: a dataframe with columns `protein1`, `protein2` and
            the scores from string
            and the extra column `r_id`

        The extra columns `s_id` and `r_id` are used to map the genes and reactions
        to the sbml_dfs. This is useful for mapping back rcpr results to the
        sbml_dfs.
    """

    dat_gene = (
        sbml_dfs.species["s_Identifiers"]
        # Get the identifiers for the given ontology
        .map(lambda ids: ids.hoist(identifier_ontology))
        .rename(COL_GENE)
        .to_frame()
        # Merge with compartmentalized species to get the sc_id
        .merge(
            sbml_dfs.compartmentalized_species[["s_id"]].reset_index(drop=False),
            left_index=True,
            right_on="s_id",
        )[[COL_GENE, "sc_id", "s_id"]]
    )

    # Perform validations
    if not dat_gene["s_id"].is_unique:
        raise ValueError("dat_gene['s_id'] must be unique")
    if not dat_gene["sc_id"].is_unique:
        raise ValueError("dat_gene['sc_id'] must be unique")
    if not dat_gene[COL_GENE].is_unique:
        raise ValueError("dat_gene[COL_GENE] must be unique")
    if dat_gene[COL_GENE].hasnans:
        raise ValueError("dat_gene[COL_GENE] must not have NaNs")

    # Reshape into the correct format
    dat_reactions = dat_gene[["sc_id", COL_GENE]].merge(
        sbml_dfs.reaction_species[["r_id", "sc_id"]], on="sc_id"
    )[[COL_GENE, "r_id"]]
    # assert that this has the correct shape, ie 2x the shape of the number
    # of reactions
    if dat_reactions.shape[0] != 2 * sbml_dfs.reactions.shape[0]:
        raise ValueError("There should be exactly 2 reactants per reaction")

    # This is the fastest way I found to reshape this into the
    # Edgelist format
    dat_reactions["flag"] = dat_reactions["r_id"].duplicated()
    dat_interactions = dat_reactions.pivot(
        index="r_id", columns="flag", values=COL_GENE
    )
    dat_interactions.columns = pd.Index([COL_PROTEIN_1, COL_PROTEIN_2], dtype=object)
    if rescale_data is not None:
        reaction_df = rescale_data(sbml_dfs.reactions_data[reaction_data])
    else:
        reaction_df = sbml_dfs.reactions_data[reaction_data]

    dat_interactions = dat_interactions.join(reaction_df).reset_index(drop=False)

    genes = pandas_to_r_dataframe(dat_gene)
    interactions = pandas_to_r_dataframe(dat_interactions)

    out = ListVector({FIELD_GENES: genes, FIELD_INTERACTIONS: interactions})
    return out


@warn_if_no_rpy2
def load_and_clean_hpa_data(
    rcpr,
    uri_hpa: str,
):
    """Load and cleans HPA data using rcpr

    Args:
        rcpr (): The rpy2 rcpr object
        uri_hpa (str): The uri of the HPA data

    Returns:
        rpy2 object: The cleaned HPA data
    """

    with NamedTemporaryFile() as f:
        # R cannot work with gcs uris
        # thus download the file to a temporary
        # location incase it is a gcs uri
        if os.path.exists(uri_hpa):
            # if the file is already a local
            # file, just use it
            path_hpa = uri_hpa
        else:
            path_hpa = f.name
            utils.copy_uri(uri_hpa, path_hpa)

        hpa_localization_data = rcpr.load_and_clean_hpa_data(path_hpa)
    return hpa_localization_data


@warn_if_no_rpy2
def load_and_clean_gtex_data(rcpr_rpy2, uri_gtex: str, by_tissue_zfpkm: bool = False):
    """Load and cleans GTEx data using rcpr

    Args:
        rcpr_rpy2 (): The rpy2 rcpr object
        uri_gtex (str): The uri of the GTEx data
        by_tissue_zfpkm (bool, optional): Whether to return the data normalized
          by tissue using zfpkm. Defaults to False.
    Returns:
        rpy2 object: The cleaned GTEx data
    """
    with NamedTemporaryFile() as f:
        # R cannot work with gcs uris
        # thus download the file to a temporary
        # location incase it is a gcs uri
        if os.path.exists(uri_gtex):
            # if the file is already a local
            # file, just use it
            path_gtex = uri_gtex
        else:
            path_gtex = f.name
            utils.copy_uri(uri_gtex, path_gtex)

        gtex_tissue_data = rcpr_rpy2.load_and_clean_gtex_data(path_gtex)

    if by_tissue_zfpkm:
        gtex_tissue_data = rcpr_rpy2.gene_expression_by_tissue(gtex_tissue_data)
    return gtex_tissue_data


def annotate_genes(
    rcpr, rcpr_graph: ListVector, data, field_name: str, **kwargs
) -> ListVector:
    """Annotates the genes in the graph with the given gene data

    See the rcpr documentation about the exact format
    required.

    Args:
        rcpr (): The rpy2 rcpr object
        rcpr_graph (ListVector): The graph to annotate
        data (complicated): "
        field_name (str): The name of the column in the gene data to annotate with

    Returns:
        ListVector: The annotated graph
    """
    # Annotate the genes
    rcpr_graph_annot = rcpr.annotate_genes(rcpr_graph, data, field_name, **kwargs)
    return rcpr_graph_annot


def trim_network_by_gene_attribute(
    rcpr,
    rcpr_graph: ListVector,
    field_name: str,
    field_value: Any = None,
    **kwargs,
) -> ListVector:
    """Trims the network by a gene attribute

    See the R function `rcpr::trim_network_by_gene_attribute` for
    more details.

    Args:
        rcpr (): The rpy2 rcpr object
        rcpr_graph (ListVector): The graph to trim
        field_name (str): The name of the column in the gene data to trim by
        field_value (Any): One or more values to trim by

    Returns:
        ListVector: The trimmed graph
    """
    if field_value is None:
        field_value = robjs.r("NaN")
    rcpr_graph_trimmed = rcpr.trim_network_by_gene_attribute(
        rcpr_graph, field_name=field_name, field_value=field_value, **kwargs
    )
    return rcpr_graph_trimmed


def apply_context_to_sbml_dfs(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    rcpr_graph: ListVector,
    inplace=True,
    remove_species=False,
) -> sbml_dfs_core.SBML_dfs:
    """Applies the context to the SBML dfs

    This is currently an in-place modification of
    the sbml_dfs object.

    Args:
        sbml_dfs (SbmlDfs): The SBML dfs to apply the context to
        rcpr_graph (ListVector): The graph to apply the context from
        inplace (bool, optional): Whether to modify the sbml_dfs in-place
            when applying the context. Defaults to True. "False" not yet implemented.
        remove_species (bool, optional): Whether to remove
            (compartmentalized) species that are no longer in the reactions.
            Defaults to False.

    Returns:
        SbmlDfs: The SBML dfs with the context applied
    """
    if not inplace:
        raise NotImplementedError("Only inplace is currently supported")

    # r_ids after trimming
    r_ids_new = set(rcpr_graph.rx("interactions")[0].rx("r_id")[0])

    # find original r_ids
    r_ids_old = set(sbml_dfs.reactions.index.tolist())

    # find the r_ids that are in the original but not in the new
    r_ids_to_remove = r_ids_old - r_ids_new

    # assert that no new r_ids were added
    if len(diff_ids := r_ids_new - r_ids_old) != 0:
        raise ValueError(
            f"New reactions present in rcpr, not present in smbl_dfs: {', '.join(diff_ids)}"
        )

    sbml_dfs.remove_reactions(r_ids_to_remove, remove_species=remove_species)

    return sbml_dfs


def sbml_dfs_to_rcpr_reactions(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    identifier_ontology: str = NETCONTEXTR_ONTOLOGY,
) -> ListVector:
    """Converts an sbml_dfs to a rcpr reaction graph

    This utility converts the sbml_dfs to the format validated by
    by `rcpr::validate_netcontextr_reactions`.

    It converts the smbl_dfs into a reaction graph by:
    - Building the `reactions` dataframe:
        - Using the `species` identifiers to map `reaction_species`
            to `genes` using the `identifier_ontology`.
            Note that one species may be split into multiple `genes`
            and multiple species may be combined into a single `gene`.
        - Converting `sbo_terms` to roles.
        - renaming `r_id` to `reaction_id`
    - Building `genes` dataframe by taking all unique `genes` from the `reactions`

    Args:
        sbml_dfs (SBML_dfs): an sbml_dfs.
        identifier_ontology (str, optional): The ontology to use for the
             identifiers. Defaults to `ensembl_gene` (default in rcpr)

    Returns:
        This is a list of dataframes that validate with validate_netcontextr_reactions:
        - `genes`: a dataframe with column `gene`
        - `reactions`: a dataframe with columns "gene", "reaction_id", "role", "rsc_id"
            representing the reaction data split up into individual reactions.
    """

    # Get the reactions
    dat_reactions = _get_reactions(sbml_dfs, identifier_ontology)
    # Get the genes
    dat_gene = dat_reactions[[COL_GENE]].drop_duplicates()
    # Note that no 1:1 mapping between genes and species can be made
    # as multiple species could have the same gene annotation
    # and also even one species could have multiple gene identifiers
    genes = pandas_to_r_dataframe(dat_gene)
    reactions = pandas_to_r_dataframe(dat_reactions)

    out = ListVector({FIELD_GENES: genes, FIELD_REACTIONS: reactions})
    return out


def trim_reactions_by_gene_attribute(
    rcpr,
    rcpr_reactions: ListVector,
    field_name: str,
    field_value: Any = None,
    **kwargs,
) -> ListVector:
    """Trims rcpr reactions by a gene attribute

    See the R function `rcpr::trim_reactions_by_gene_attribute` for
    more details.

    Args:
        rcpr (): The rpy2 rcpr object
        rcpr_reactions (ListVector): The graph to trim
        field_name (str): The name of the column in the gene data to trim by
        field_value (Any): One or more values to trim by

    Returns:
        ListVector: The trimmed graph
    """
    if field_value is None:
        field_value = robjs.r("NaN")
    rcpr_reactions_trimmed = rcpr.trim_reactions_by_gene_attribute(
        rcpr_reactions, field_name=field_name, field_value=field_value, **kwargs
    )
    return rcpr_reactions_trimmed


def apply_reactions_context_to_sbml_dfs(
    sbml_dfs: sbml_dfs_core.SBML_dfs,
    rcpr_reactions: ListVector,
    considered_reactions: Iterable[str] | None = None,
    inplace=True,
    remove_species=False,
) -> sbml_dfs_core.SBML_dfs:
    """Applies the context to the SBML dfs

    This is currently an in-place modification of
    the sbml_dfs object.

    Args:
        sbml_dfs (sbml_dfs_core.SBML_dfs): The SBML dfs to apply the context to
        rcpr_reactions (ListVector): The contextualized
        considered_reactions (Iterable[str], optional): The reactions that were
            considered for contextualisation. If None, all reactions that are
            in the sbml_dfs are considered and filtered out if they are not part of
            the rcpr_reactions. If provided, only reactions considered and not part
            of the rcpr_reactions are removed. Defaults to None.
        inplace (bool, optional): Whether to apply the context inplace.
            Only True currently implemented.
        remove_species (bool, optional): Whether to remove
            (compartmentalized) species that are no longer in the reactions.
            Defaults to False.

    Returns:
        SbmlDfs: The SBML dfs with the context applied
    """
    if not inplace:
        raise NotImplementedError("Only inplace is currently supported")

    # r_ids after trimming
    r_ids_new = _get_rids_from_rcpr_reactions(rcpr_reactions)

    # find original r_ids
    if considered_reactions is None:
        r_ids_old = set(sbml_dfs.reactions.index.tolist())
    else:
        r_ids_old = set(considered_reactions)

    # find the r_ids that are in the original but not in the new
    r_ids_to_remove = r_ids_old - r_ids_new

    # assert that no new r_ids were added
    if len(diff_ids := r_ids_new - r_ids_old) != 0:
        raise ValueError(
            "New reactions present in rcpr, not present in the considered "
            f"reactions: {', '.join(diff_ids)}"
        )

    sbml_dfs.remove_reactions(r_ids_to_remove, remove_species=remove_species)

    return sbml_dfs


def _get_rids_from_rcpr_reactions(rcpr_reactions: ListVector) -> set[str]:
    """Gets the r_ids from the rcpr reactions"""
    return set(rcpr_reactions.rx(FIELD_REACTIONS)[0].rx(COL_REACTION_ID)[0])


def _get_reactions(
    sbml_dfs: sbml_dfs_core.SBML_dfs, identifier_ontology: str = NETCONTEXTR_ONTOLOGY
) -> pd.DataFrame:
    """Gets the reactions from the sbml_dfs"""
    dat_reaction = (
        sbml_dfs.species["s_Identifiers"]
        # Get the identifiers for the given ontology
        .map(lambda ids: ids.hoist(identifier_ontology, squeeze=False))
        .map(lambda x: x if len(x) > 0 else None)
        .dropna()
        .rename(COL_GENE)
        .to_frame()
        .explode(COL_GENE)
        # Merge with compartmentalized species to get the sc_id
        .merge(
            sbml_dfs.compartmentalized_species[["s_id"]].reset_index(drop=False),
            left_index=True,
            right_on="s_id",
        )[[COL_GENE, "sc_id"]]
        .merge(
            sbml_dfs.reaction_species[
                ["r_id", "sc_id", "sbo_term", "stoichiometry"]
            ].reset_index(drop=False),
            on="sc_id",
        )
        .assign(**{COL_ROLE: lambda x: x["sbo_term"].map(NETCONTEXTR_SBO_MAP)})
        .rename({"r_id": COL_REACTION_ID, "stoichiometry": COL_STOICHIOMETRY}, axis=1)
    )
    fil = dat_reaction[COL_ROLE].isna()
    if fil.sum() > 0:
        missing_sbo_terms = dat_reaction.loc[fil, "sbo_term"].unique()
        logger.warning(
            f"Found {fil.sum()} reactions had an sbo term that was not"
            "mappable to a rcpr role. These are ignored. "
            f"The sbo terms are: {', '.join(missing_sbo_terms)}"
        )

        dat_reaction = dat_reaction.loc[~fil, :]
    return dat_reaction[
        [COL_ROLE, COL_GENE, COL_REACTION_ID, COL_STOICHIOMETRY, "rsc_id"]
    ]



from __future__ import annotations

import os

import numpy as np
import pandas as pd
import pytest
from napistu import sbml_dfs_core
from napistu.ingestion import sbml
from napistu.modify import pathwayannot

from napistu import identifiers as napistu_identifiers
from napistu.constants import SBML_DFS
from napistu.sbml_dfs_core import SBML_dfs


def test_drop_cofactors(sbml_dfs):
    starting_rscs = sbml_dfs.reaction_species.shape[0]
    reduced_dfs = pathwayannot.drop_cofactors(sbml_dfs)

    assert starting_rscs - reduced_dfs.reaction_species.shape[0] == 20


def test_sbml_dfs_from_dict_required(sbml_dfs):
    val_dict = {k: getattr(sbml_dfs, k) for k in sbml_dfs._required_entities}
    sbml_dfs2 = sbml_dfs_core.SBML_dfs(val_dict)
    sbml_dfs2.validate()

    for k in sbml_dfs._required_entities:
        assert getattr(sbml_dfs2, k).equals(getattr(sbml_dfs, k))


def test_sbml_dfs_species_data(sbml_dfs):
    data = pd.DataFrame({"bla": [1, 2, 3]}, index=sbml_dfs.species.iloc[:3].index)
    sbml_dfs.add_species_data("test", data)
    sbml_dfs.validate()


def test_sbml_dfs_species_data_existing(sbml_dfs):
    data = pd.DataFrame({"bla": [1, 2, 3]}, index=sbml_dfs.species.iloc[:3].index)
    sbml_dfs.add_species_data("test", data)
    with pytest.raises(ValueError):
        sbml_dfs.add_species_data("test", data)


def test_sbml_dfs_species_data_validation(sbml_dfs):
    data = pd.DataFrame({"bla": [1, 2, 3]})
    sbml_dfs.species_data["test"] = data
    with pytest.raises(ValueError):
        sbml_dfs.validate()


def test_sbml_dfs_species_data_missing_idx(sbml_dfs):
    data = pd.DataFrame({"bla": [1, 2, 3]})
    with pytest.raises(ValueError):
        sbml_dfs.add_species_data("test", data)


def test_sbml_dfs_species_data_duplicated_idx(sbml_dfs):
    an_s_id = sbml_dfs.species.iloc[0].index[0]
    dup_idx = pd.Series([an_s_id, an_s_id], name="s_id")
    data = pd.DataFrame({"bla": [1, 2]}, index=dup_idx)

    with pytest.raises(ValueError):
        sbml_dfs.add_species_data("test", data)


def test_sbml_dfs_species_data_wrong_idx(sbml_dfs):
    data = pd.DataFrame(
        {"bla": [1, 2, 3]}, index=pd.Series(["bla1", "bla2", "bla3"], name="s_id")
    )
    with pytest.raises(ValueError):
        sbml_dfs.add_species_data("test", data)


def test_sbml_dfs_reactions_data(sbml_dfs):
    reactions_data = pd.DataFrame(
        {"bla": [1, 2, 3]}, index=sbml_dfs.reactions.iloc[:3].index
    )
    sbml_dfs.add_reactions_data("test", reactions_data)
    sbml_dfs.validate()


def test_sbml_dfs_reactions_data_existing(sbml_dfs):
    reactions_data = pd.DataFrame(
        {"bla": [1, 2, 3]}, index=sbml_dfs.reactions.iloc[:3].index
    )
    sbml_dfs.add_reactions_data("test", reactions_data)
    with pytest.raises(ValueError):
        sbml_dfs.add_reactions_data("test", reactions_data)


def test_sbml_dfs_reactions_data_validate(sbml_dfs):
    data = pd.DataFrame({"bla": [1, 2, 3]})
    sbml_dfs.reactions_data["test"] = data
    with pytest.raises(ValueError):
        sbml_dfs.validate()


def test_sbml_dfs_reactions_data_missing_idx(sbml_dfs):
    data = pd.DataFrame({"bla": [1, 2, 3]})
    with pytest.raises(ValueError):
        sbml_dfs.add_reactions_data("test", data)


def test_sbml_dfs_reactions_data_duplicated_idx(sbml_dfs):
    an_r_id = sbml_dfs.reactions.iloc[0].index[0]
    dup_idx = pd.Series([an_r_id, an_r_id], name="r_id")
    data = pd.DataFrame({"bla": [1, 2]}, index=dup_idx)
    with pytest.raises(ValueError):
        sbml_dfs.add_reactions_data("test", data)


def test_sbml_dfs_reactions_data_wrong_idx(sbml_dfs):
    data = pd.DataFrame(
        {"bla": [1, 2, 3]}, index=pd.Series(["bla1", "bla2", "bla3"], name="r_id")
    )
    with pytest.raises(ValueError):
        sbml_dfs.add_reactions_data("test", data)


def test_sbml_dfs_remove_species_check_species(sbml_dfs):
    s_id = [sbml_dfs.species.index[0]]
    sbml_dfs._remove_species(s_id)
    assert s_id[0] not in sbml_dfs.species.index
    sbml_dfs.validate()


def test_sbml_dfs_remove_species_check_cspecies(sbml_dfs):
    s_id = [sbml_dfs.compartmentalized_species["s_id"].iloc[0]]
    sbml_dfs._remove_species(s_id)
    assert s_id[0] not in sbml_dfs.compartmentalized_species.index
    sbml_dfs.validate()


@pytest.fixture
def sbml_dfs_w_data(sbml_dfs):
    sbml_dfs.add_species_data(
        "test_species",
        pd.DataFrame({"test1": [1, 2]}, index=sbml_dfs.species.index[:2]),
    )
    sbml_dfs.add_reactions_data(
        "test_reactions",
        pd.DataFrame({"test2": [1, 2, 3]}, index=sbml_dfs.reactions.index[:3]),
    )
    return sbml_dfs


def test_sbml_dfs_remove_species_check_data(sbml_dfs_w_data):
    data = list(sbml_dfs_w_data.species_data.values())[0]
    s_id = [data.index[0]]
    sbml_dfs_w_data._remove_species(s_id)
    data_2 = list(sbml_dfs_w_data.species_data.values())[0]
    assert s_id[0] not in data_2.index
    sbml_dfs_w_data.validate()


def test_sbml_dfs_remove_cspecies_check_cspecies(sbml_dfs):
    s_id = [sbml_dfs.compartmentalized_species.index[0]]
    sbml_dfs._remove_compartmentalized_species(s_id)
    assert s_id[0] not in sbml_dfs.compartmentalized_species.index
    sbml_dfs.validate()


def test_sbml_dfs_remove_cspecies_check_reaction_species(sbml_dfs):
    sc_id = [sbml_dfs.reaction_species["sc_id"].iloc[0]]
    sbml_dfs._remove_compartmentalized_species(sc_id)
    assert sc_id[0] not in sbml_dfs.reaction_species["sc_id"]
    sbml_dfs.validate()


def test_sbml_dfs_remove_reactions_check_reactions(sbml_dfs):
    r_id = [sbml_dfs.reactions.index[0]]
    sbml_dfs.remove_reactions(r_id)
    assert r_id[0] not in sbml_dfs.reactions.index
    sbml_dfs.validate()


def test_sbml_dfs_remove_reactions_check_reaction_species(sbml_dfs):
    r_id = [sbml_dfs.reaction_species["r_id"].iloc[0]]
    sbml_dfs.remove_reactions(r_id)
    assert r_id[0] not in sbml_dfs.reaction_species["r_id"]
    sbml_dfs.validate()


def test_sbml_dfs_remove_reactions_check_data(sbml_dfs_w_data):
    data = list(sbml_dfs_w_data.reactions_data.values())[0]
    r_id = [data.index[0]]
    sbml_dfs_w_data.remove_reactions(r_id)
    data_2 = list(sbml_dfs_w_data.reactions_data.values())[0]
    assert r_id[0] not in data_2.index
    sbml_dfs_w_data.validate()


def test_sbml_dfs_remove_reactions_check_species(sbml_dfs):
    # find all r_ids for a species and check if
    # removing all these reactions also removes the species
    s_id = sbml_dfs.species.index[0]
    dat = (
        sbml_dfs.compartmentalized_species.query("s_id == @s_id").merge(
            sbml_dfs.reaction_species, on="sc_id"
        )
    )[["r_id", "sc_id"]]
    r_ids = dat["r_id"]
    sc_ids = dat["sc_id"]
    sbml_dfs.remove_reactions(r_ids, remove_species=True)
    for sc_id in sc_ids:
        assert sc_id not in sbml_dfs.compartmentalized_species.index
    assert s_id not in sbml_dfs.species.index


def test_formula(sbml_dfs):
    # create a formula string

    an_r_id = sbml_dfs.reactions.index[0]

    reaction_species_df = sbml_dfs.reaction_species[
        sbml_dfs.reaction_species["r_id"] == an_r_id
    ].merge(sbml_dfs.compartmentalized_species, left_on="sc_id", right_index=True)

    formula_str = sbml_dfs_core.construct_formula_string(
        reaction_species_df, sbml_dfs.reactions, name_var="sc_name"
    )

    assert isinstance(formula_str, str)
    assert (
        formula_str
        == "CO2 [extracellular region] -> CO2 [cytosol] ---- modifiers: AQP1 tetramer [plasma membrane]]"
    )


def test_read_sbml_with_invalid_ids():
    SBML_W_BAD_IDS = "R-HSA-166658.sbml"
    test_path = os.path.abspath(os.path.join(__file__, os.pardir))
    sbml_w_bad_ids_path = os.path.join(test_path, "test_data", SBML_W_BAD_IDS)
    assert os.path.isfile(sbml_w_bad_ids_path)

    # invalid identifiers still create a valid sbml_dfs
    sbml_w_bad_ids = sbml.SBML(sbml_w_bad_ids_path)
    assert isinstance(sbml_dfs_core.SBML_dfs(sbml_w_bad_ids), sbml_dfs_core.SBML_dfs)


def test_stubbed_compartment():
    compartment = sbml_dfs_core._stub_compartments()

    assert compartment["c_Identifiers"].iloc[0].ids[0] == {
        "ontology": "go",
        "identifier": "GO:0005575",
        "url": "https://www.ebi.ac.uk/QuickGO/term/GO:0005575",
        "bqb": "BQB_IS",
    }


def test_get_table(sbml_dfs):
    assert isinstance(sbml_dfs.get_table("species"), pd.DataFrame)
    assert isinstance(sbml_dfs.get_table("species", {"id"}), pd.DataFrame)

    # invalid table
    with pytest.raises(ValueError):
        sbml_dfs.get_table("foo", {"id"})

    # bad type
    with pytest.raises(TypeError):
        sbml_dfs.get_table("reaction_species", "id")

    # reaction species don't have ids
    with pytest.raises(ValueError):
        sbml_dfs.get_table("reaction_species", {"id"})


def test_search_by_name(sbml_dfs_metabolism):
    assert sbml_dfs_metabolism.search_by_name("atp", "species", False).shape[0] == 1
    assert sbml_dfs_metabolism.search_by_name("pyr", "species").shape[0] == 3
    assert sbml_dfs_metabolism.search_by_name("kinase", "reactions").shape[0] == 4


def test_search_by_id(sbml_dfs_metabolism):
    identifiers_tbl = sbml_dfs_metabolism.get_identifiers("species")
    ids, species = sbml_dfs_metabolism.search_by_ids(
        ["P40926"], "species", identifiers_tbl
    )
    assert ids.shape[0] == 1
    assert species.shape[0] == 1

    ids, species = sbml_dfs_metabolism.search_by_ids(
        ["57540", "30744"], "species", identifiers_tbl, {"chebi"}
    )
    assert ids.shape[0] == 2
    assert species.shape[0] == 2

    ids, species = sbml_dfs_metabolism.search_by_ids(
        ["baz"], "species", identifiers_tbl
    )
    assert ids.shape[0] == 0
    assert species.shape[0] == 0


def test_species_status(sbml_dfs):

    species = sbml_dfs.species
    select_species = species[species["s_name"] == "OxyHbA"]
    assert select_species.shape[0] == 1

    status = sbml_dfs_core.species_status(select_species.index[0], sbml_dfs)
    assert (
        status["r_formula_str"][0]
        == "4.0 H+ + OxyHbA + 4.0 CO2 -> 4.0 O2 + Protonated Carbamino DeoxyHbA [cytosol]"
    )


def test_get_identifiers_handles_missing_values():

    # Minimal DataFrame with all types
    df = pd.DataFrame(
        {
            SBML_DFS.S_NAME: ["A", "B", "C", "D"],
            SBML_DFS.S_IDENTIFIERS: [
                napistu_identifiers.Identifiers([]),
                None,
                np.nan,
                pd.NA,
            ],
            SBML_DFS.S_SOURCE: [None, None, None, None],
        },
        index=["s1", "s2", "s3", "s4"],
    )
    df.index.name = SBML_DFS.S_ID

    sbml_dict = {
        SBML_DFS.COMPARTMENTS: pd.DataFrame(
            {
                SBML_DFS.C_NAME: ["cytosol"],
                SBML_DFS.C_IDENTIFIERS: [None],
                SBML_DFS.C_SOURCE: [None],
            },
            index=["c1"],
        ),
        SBML_DFS.SPECIES: df,
        SBML_DFS.COMPARTMENTALIZED_SPECIES: pd.DataFrame(
            {
                SBML_DFS.SC_NAME: ["A [cytosol]"],
                SBML_DFS.S_ID: ["s1"],
                SBML_DFS.C_ID: ["c1"],
                SBML_DFS.SC_SOURCE: [None],
            },
            index=["sc1"],
        ),
        SBML_DFS.REACTIONS: pd.DataFrame(
            {
                SBML_DFS.R_NAME: [],
                SBML_DFS.R_IDENTIFIERS: [],
                SBML_DFS.R_SOURCE: [],
                SBML_DFS.R_ISREVERSIBLE: [],
            },
            index=[],
        ),
        SBML_DFS.REACTION_SPECIES: pd.DataFrame(
            {
                SBML_DFS.R_ID: [],
                SBML_DFS.SC_ID: [],
                SBML_DFS.STOICHIOMETRY: [],
                SBML_DFS.SBO_TERM: [],
            },
            index=[],
        ),
    }
    sbml = SBML_dfs(sbml_dict, validate=False)
    result = sbml.get_identifiers(SBML_DFS.SPECIES)
    assert result.shape[0] == 0 or all(
        result[SBML_DFS.S_ID] == "s1"
    ), "Only Identifiers objects should be returned."



"""
Core MCP server implementation for Napistu.
"""

import logging

from mcp.server import FastMCP

from napistu.mcp import codebase
from napistu.mcp import documentation
from napistu.mcp import execution
from napistu.mcp import tutorials

from napistu.mcp.profiles import ServerProfile

logger = logging.getLogger(__name__)


def create_server(profile: ServerProfile, **kwargs) -> FastMCP:
    """
    Create an MCP server based on a profile configuration.

    Parameters
    ----------
    profile : ServerProfile
        Server profile to use. All configuration must be set in the profile.
    **kwargs
        Additional arguments to pass to the FastMCP constructor such as host and port.

    Returns
    -------
    FastMCP
        Configured FastMCP server instance.
    """

    config = profile.get_config()

    # Create the server with FastMCP-specific parameters
    # Pass all kwargs directly to the FastMCP constructor
    mcp = FastMCP(config["server_name"], **kwargs)

    if config["enable_documentation"]:
        logger.info("Registering documentation components")
        documentation.register_components(mcp)
    if config["enable_codebase"]:
        logger.info("Registering codebase components")
        codebase.register_components(mcp)
    if config["enable_execution"]:
        logger.info("Registering execution components")
        execution.register_components(
            mcp,
            session_context=config["session_context"],
            object_registry=config["object_registry"],
        )
    if config["enable_tutorials"]:
        logger.info("Registering tutorials components")
        tutorials.register_components(mcp)
    return mcp


async def initialize_components(profile: ServerProfile) -> None:
    """
    Asynchronously initialize all enabled components for the MCP server, using the provided ServerProfile.

    Parameters
    ----------
    profile : ServerProfile
        The profile whose configuration determines which components to initialize.

    Returns
    -------
    None
    """
    config = profile.get_config()
    if config["enable_documentation"]:
        logger.info("Initializing documentation components")
        await documentation.initialize_components()
    if config["enable_codebase"]:
        logger.info("Initializing codebase components")
        await codebase.initialize_components()
    if config["enable_tutorials"]:
        logger.info("Initializing tutorials components")
        await tutorials.initialize_components()
    if config["enable_execution"]:
        logger.info("Initializing execution components")
        await execution.initialize_components()



from __future__ import annotations

import os

import pandas as pd
import pytest
from napistu import indices

test_path = os.path.abspath(os.path.join(__file__, os.pardir))
test_data = os.path.join(test_path, "test_data")


def test_pwindex_from_file():
    pw_index_path = os.path.join(test_data, "pw_index.tsv")
    pw_index = indices.PWIndex(pw_index_path)

    assert pw_index.index.shape == (5, 6)


def test_pwindex_from_df():
    stub_pw_df = pd.DataFrame(
        {
            "file": "DNE",
            "source": "The farm",
            "species": "Gallus gallus",
            "pathway_id": "chickens",
            "name": "Chickens",
            "date": "2020-01-01",
        },
        index=[0],
    )

    assert indices.PWIndex(pw_index=stub_pw_df, validate_paths=False).index.equals(
        stub_pw_df
    )

    with pytest.raises(FileNotFoundError) as _:
        indices.PWIndex(pw_index=stub_pw_df, pw_index_base_path="missing_directory")

    with pytest.raises(FileNotFoundError) as _:
        indices.PWIndex(pw_index=stub_pw_df, pw_index_base_path=test_data)


@pytest.fixture
def pw_testindex():
    pw_index = indices.PWIndex(os.path.join(test_data, "pw_index.tsv"))
    return pw_index


def test_index(pw_testindex):
    pw_index = pw_testindex
    full_index_shape = (5, 6)
    assert pw_index.index.shape == full_index_shape

    ref_index = pw_index.index.copy()
    pw_index.filter(sources="Reactome")
    assert pw_index.index.shape == full_index_shape

    pw_index.filter(species="Homo sapiens")
    assert pw_index.index.shape == full_index_shape

    pw_index.filter(sources=("Reactome",))
    assert pw_index.index.shape == full_index_shape

    pw_index.filter(species=("Homo sapiens",))
    assert pw_index.index.shape == full_index_shape

    pw_index.filter(sources="NotValid")
    assert pw_index.index.shape == (0, full_index_shape[1])
    pw_index.index = ref_index.copy()

    pw_index.filter(species="NotValid")
    assert pw_index.index.shape == (0, full_index_shape[1])
    pw_index.index = ref_index.copy()

    pw_index.search("erythrocytes")
    assert pw_index.index.shape == (2, 6)
    pw_index.index = ref_index.copy()

    pw_index.search("erythrocytes|HYDROCARBON")
    assert pw_index.index.shape == (3, 6)


def test_missing_file(pw_testindex):
    pw_index = pw_testindex
    pw_index.index.loc[0, "file"] = "not_existing.sbml"
    with pytest.raises(FileNotFoundError) as _:
        pw_index._check_files()



from __future__ import annotations

import os
import pytest

import pandas as pd
from napistu import sbml_dfs_core
from napistu.constants import DEFAULT_WT_TRANS
from napistu.constants import MINI_SBO_FROM_NAME
from napistu.ingestion import sbml
from napistu.network import neighborhoods
from napistu.network import net_create
from napistu.network import net_utils
from napistu.network import paths

test_path = os.path.abspath(os.path.join(__file__, os.pardir))
test_data = os.path.join(test_path, "test_data")

sbml_path = os.path.join(test_data, "R-HSA-1237044.sbml")
sbml_model = sbml.SBML(sbml_path).model
sbml_dfs = sbml_dfs_core.SBML_dfs(sbml_model)

# create a dict containing reaction species for a few example reactions
reaction_species_examples_dict = dict()

# stub with a random reaction
r_id = sbml_dfs.reactions.index[0]

reaction_species_examples_dict["valid_interactor"] = pd.DataFrame(
    {
        "r_id": [r_id, r_id],
        "sbo_term": [
            MINI_SBO_FROM_NAME["interactor"],
            MINI_SBO_FROM_NAME["interactor"],
        ],
        "sc_id": ["sc1", "sc2"],
        "stoichiometry": [0, 0],
    }
).set_index(["r_id", "sbo_term"])

reaction_species_examples_dict["invalid_interactor"] = pd.DataFrame(
    {
        "r_id": [r_id, r_id],
        "sbo_term": [
            MINI_SBO_FROM_NAME["interactor"],
            MINI_SBO_FROM_NAME["product"],
        ],
        "sc_id": ["sc1", "sc2"],
        "stoichiometry": [0, 0],
    }
).set_index(["r_id", "sbo_term"])


# simple reaction with just substrates and products
reaction_species_examples_dict["sub_and_prod"] = pd.DataFrame(
    {
        "r_id": [r_id, r_id],
        "sbo_term": [MINI_SBO_FROM_NAME["reactant"], MINI_SBO_FROM_NAME["product"]],
        "sc_id": ["sub", "prod"],
        "stoichiometry": [-1, 1],
    }
).set_index(["r_id", "sbo_term"])

reaction_species_examples_dict["stimulator"] = pd.DataFrame(
    {
        "r_id": [r_id, r_id, r_id],
        "sbo_term": [
            MINI_SBO_FROM_NAME["reactant"],
            MINI_SBO_FROM_NAME["product"],
            MINI_SBO_FROM_NAME["stimulator"],
        ],
        "sc_id": ["sub", "prod", "stim"],
        "stoichiometry": [-1, 1, 0],
    }
).set_index(["r_id", "sbo_term"])

reaction_species_examples_dict["all_entities"] = pd.DataFrame(
    {
        "r_id": [r_id, r_id, r_id, r_id],
        "sbo_term": [
            MINI_SBO_FROM_NAME["reactant"],
            MINI_SBO_FROM_NAME["product"],
            MINI_SBO_FROM_NAME["stimulator"],
            MINI_SBO_FROM_NAME["catalyst"],
        ],
        "sc_id": ["sub", "prod", "stim", "cat"],
        "stoichiometry": [-1, 1, 0, 0],
    }
).set_index(["r_id", "sbo_term"])

reaction_species_examples_dict["no_substrate"] = pd.DataFrame(
    {
        "r_id": [r_id, r_id, r_id, r_id, r_id],
        "sbo_term": [
            MINI_SBO_FROM_NAME["product"],
            MINI_SBO_FROM_NAME["stimulator"],
            MINI_SBO_FROM_NAME["stimulator"],
            MINI_SBO_FROM_NAME["inhibitor"],
            MINI_SBO_FROM_NAME["catalyst"],
        ],
        "sc_id": ["prod", "stim1", "stim2", "inh", "cat"],
        "stoichiometry": [1, 0, 0, 0, 0],
    }
).set_index(["r_id", "sbo_term"])


def test_create_cpr_graph():
    _ = net_create.create_cpr_graph(sbml_dfs, graph_type="bipartite")
    _ = net_create.create_cpr_graph(sbml_dfs, graph_type="regulatory")
    _ = net_create.create_cpr_graph(sbml_dfs, graph_type="surrogate")


def test_create_cpr_graph_none_attrs():
    # Should not raise when reaction_graph_attrs is None
    _ = net_create.create_cpr_graph(
        sbml_dfs, reaction_graph_attrs=None, graph_type="bipartite"
    )


def test_igraph_construction():
    _ = net_create.process_cpr_graph(sbml_dfs)


def test_process_cpr_graph_none_attrs():
    # Should not raise when reaction_graph_attrs is None
    _ = net_create.process_cpr_graph(sbml_dfs, reaction_graph_attrs=None)


@pytest.mark.skip_on_windows
def test_igraph_loading():
    # test read/write of an igraph network
    directeds = [True, False]
    graph_types = ["bipartite", "regulatory"]

    net_utils.export_networks(
        sbml_dfs,
        model_prefix="tmp",
        outdir="/tmp",
        directeds=directeds,
        graph_types=graph_types,
    )

    for graph_type in graph_types:
        for directed in directeds:
            import_pkl_path = net_utils._create_network_save_string(
                model_prefix="tmp",
                outdir="/tmp",
                directed=directed,
                graph_type=graph_type,
            )
            network_graph = net_utils.read_network_pkl(
                model_prefix="tmp",
                network_dir="/tmp",
                directed=directed,
                graph_type=graph_type,
            )

            assert network_graph.is_directed() == directed
            # cleanup
            os.unlink(import_pkl_path)


def test_shortest_paths():
    species = sbml_dfs.species
    source_species = species[species["s_name"] == "NADH"]
    dest_species = species[species["s_name"] == "NAD+"]
    target_species_paths = net_utils.compartmentalize_species_pairs(
        sbml_dfs, source_species.index.tolist(), dest_species.index.tolist()
    )

    # directed graph
    cpr_graph = net_create.process_cpr_graph(
        sbml_dfs, directed=True, weighting_strategy="topology"
    )
    (
        all_shortest_reaction_paths_df,
        all_shortest_reaction_path_edges_df,
        edge_sources,
        paths_graph,
    ) = paths.find_all_shortest_reaction_paths(
        cpr_graph, sbml_dfs, target_species_paths, weight_var="weights"
    )

    # undirected graph
    cpr_graph = net_create.process_cpr_graph(
        sbml_dfs, directed=False, weighting_strategy="topology"
    )
    (
        all_shortest_reaction_paths_df,
        all_shortest_reaction_path_edges_df,
        edge_sources,
        paths_graph,
    ) = paths.find_all_shortest_reaction_paths(
        cpr_graph, sbml_dfs, target_species_paths, weight_var="weights"
    )

    assert all_shortest_reaction_paths_df.shape[0] == 3


def test_neighborhood():
    species = sbml_dfs.species
    source_species = species[species["s_name"] == "NADH"].index.tolist()

    query_sc_species = net_utils.compartmentalize_species(sbml_dfs, source_species)
    compartmentalized_species = query_sc_species["sc_id"].tolist()

    cpr_graph = net_create.process_cpr_graph(
        sbml_dfs, directed=True, weighting_strategy="topology"
    )

    neighborhood = neighborhoods.find_neighborhoods(
        sbml_dfs,
        cpr_graph,
        compartmentalized_species=compartmentalized_species,
        order=3,
    )

    assert neighborhood["species_73473"]["vertices"].shape[0] == 6


def test_format_interactors():
    # interactions are formatted

    graph_hierarchy_df = net_create._create_graph_hierarchy_df("regulatory")

    assert (
        net_create._format_tiered_reaction_species(
            r_id,
            reaction_species_examples_dict["valid_interactor"],
            sbml_dfs,
            graph_hierarchy_df,
        ).shape[0]
        == 1
    )

    print("Re-enable test once Issue #102 is solved")

    # catch error from invalid interactor specification
    # with pytest.raises(ValueError) as excinfo:
    #    net_create._format_tiered_reaction_species(
    #        r_id, reaction_species_examples_dict["invalid_interactor"], sbml_dfs
    #    )
    # assert str(excinfo.value).startswith("Invalid combinations of SBO_terms")

    # simple reaction with just substrates and products
    assert (
        net_create._format_tiered_reaction_species(
            r_id,
            reaction_species_examples_dict["sub_and_prod"],
            sbml_dfs,
            graph_hierarchy_df,
        ).shape[0]
        == 2
    )

    # add a stimulator (activator)
    rxn_edges = net_create._format_tiered_reaction_species(
        r_id, reaction_species_examples_dict["stimulator"], sbml_dfs, graph_hierarchy_df
    )

    assert rxn_edges.shape[0] == 3
    assert rxn_edges.iloc[0][["from", "to"]].tolist() == ["stim", "sub"]

    # add catalyst + stimulator
    rxn_edges = net_create._format_tiered_reaction_species(
        r_id,
        reaction_species_examples_dict["all_entities"],
        sbml_dfs,
        graph_hierarchy_df,
    )

    assert rxn_edges.shape[0] == 4
    assert rxn_edges.iloc[0][["from", "to"]].tolist() == ["stim", "cat"]
    assert rxn_edges.iloc[1][["from", "to"]].tolist() == ["cat", "sub"]

    # no substrate
    rxn_edges = net_create._format_tiered_reaction_species(
        r_id,
        reaction_species_examples_dict["no_substrate"],
        sbml_dfs,
        graph_hierarchy_df,
    )

    assert rxn_edges.shape[0] == 5
    # stimulator -> reactant
    assert rxn_edges.iloc[0][["from", "to"]].tolist() == ["stim1", "cat"]
    assert rxn_edges.iloc[1][["from", "to"]].tolist() == ["stim2", "cat"]
    assert rxn_edges.iloc[2][["from", "to"]].tolist() == ["inh", "cat"]

    # use the surrogate model tiered layout also

    graph_hierarchy_df = net_create._create_graph_hierarchy_df("surrogate")

    rxn_edges = net_create._format_tiered_reaction_species(
        r_id,
        reaction_species_examples_dict["all_entities"],
        sbml_dfs,
        graph_hierarchy_df,
    )

    assert rxn_edges.shape[0] == 4
    assert rxn_edges.iloc[0][["from", "to"]].tolist() == ["stim", "sub"]
    assert rxn_edges.iloc[1][["from", "to"]].tolist() == ["sub", "cat"]


def test_reverse_network_edges():
    graph_hierarchy_df = net_create._create_graph_hierarchy_df("regulatory")

    rxn_edges = net_create._format_tiered_reaction_species(
        r_id,
        reaction_species_examples_dict["all_entities"],
        sbml_dfs,
        graph_hierarchy_df,
    )
    augmented_network_edges = rxn_edges.assign(r_isreversible=True)
    augmented_network_edges["sc_parents"] = range(0, augmented_network_edges.shape[0])
    augmented_network_edges["sc_children"] = range(
        augmented_network_edges.shape[0], 0, -1
    )

    assert net_create._reverse_network_edges(augmented_network_edges).shape[0] == 2


def test_net_polarity():
    polarity_series = pd.Series(
        ["ambiguous", "ambiguous"], index=[0, 1], name="link_polarity"
    )
    assert all(
        [x == "ambiguous" for x in paths._calculate_net_polarity(polarity_series)]
    )

    polarity_series = pd.Series(
        ["activation", "inhibition", "inhibition", "ambiguous"],
        index=range(0, 4),
        name="link_polarity",
    )
    assert paths._calculate_net_polarity(polarity_series) == [
        "activation",
        "inhibition",
        "activation",
        "ambiguous activation",
    ]
    assert paths._terminal_net_polarity(polarity_series) == "ambiguous activation"


def test_entity_validation():
    entity_attrs = {"table": "reactions", "variable": "foo"}

    assert net_create._EntityAttrValidator(**entity_attrs).model_dump() == {
        **entity_attrs,
        **{"trans": DEFAULT_WT_TRANS},
    }


################################################
# __main__
################################################

if __name__ == "__main__":
    test_create_cpr_graph()
    test_igraph_loading()
    test_igraph_construction()
    test_shortest_paths()
    test_neighborhood()
    test_format_interactors()
    test_reverse_network_edges()
    test_entity_validation()



"""
Codebase exploration components for the Napistu MCP server.
"""

from napistu.mcp.constants import NAPISTU_PY_READTHEDOCS_API

from fastmcp import FastMCP

from typing import Dict, Any
import json

from napistu.mcp import codebase_utils
from napistu.mcp import utils as mcp_utils

# Global cache for codebase information
_codebase_cache = {
    "modules": {},
    "classes": {},
    "functions": {},
}


async def initialize_components() -> bool:
    """
    Initialize codebase components.

    Returns
    -------
    bool
        True if initialization is successful.
    """
    global _codebase_cache
    # Load documentation from the ReadTheDocs API
    _codebase_cache["modules"] = await codebase_utils.read_read_the_docs(
        NAPISTU_PY_READTHEDOCS_API
    )
    # Extract functions and classes from the modules
    _codebase_cache["functions"], _codebase_cache["classes"] = (
        codebase_utils.extract_functions_and_classes_from_modules(
            _codebase_cache["modules"]
        )
    )
    return True


def register_components(mcp: FastMCP):
    """
    Register codebase exploration components with the MCP server.

    Args:
        mcp: FastMCP server instance
    """
    global _codebase_cache

    # Register resources
    @mcp.resource("napistu://codebase/summary")
    async def get_codebase_summary() -> Dict[str, Any]:
        """
        Get a summary of the Napistu codebase structure.
        """
        return {
            "modules": list(_codebase_cache["modules"].keys()),
            "top_level_classes": [
                class_name
                for class_name, info in _codebase_cache["classes"].items()
                if "." not in class_name  # Only include top-level classes
            ],
            "top_level_functions": [
                func_name
                for func_name, info in _codebase_cache["functions"].items()
                if "." not in func_name  # Only include top-level functions
            ],
        }

    @mcp.resource("napistu://codebase/modules/{module_name}")
    async def get_module_details(module_name: str) -> Dict[str, Any]:
        """
        Get detailed information about a specific module.

        Args:
            module_name: Name of the module
        """
        if module_name not in _codebase_cache["modules"]:
            return {"error": f"Module {module_name} not found"}

        return _codebase_cache["modules"][module_name]

    # Register tools
    @mcp.tool()
    async def search_codebase(query: str) -> Dict[str, Any]:
        """
        Search the codebase for a specific query.

        Args:
            query: Search term

        Returns:
            Dictionary with search results organized by code element type, including snippets for context.
        """
        results = {
            "modules": [],
            "classes": [],
            "functions": [],
        }

        # Search modules
        for module_name, info in _codebase_cache["modules"].items():
            # Use docstring or description for snippet
            doc = info.get("doc") or info.get("description") or ""
            module_text = json.dumps(info)
            if query.lower() in module_text.lower():
                snippet = mcp_utils.get_snippet(doc, query)
                results["modules"].append(
                    {
                        "name": module_name,
                        "description": doc,
                        "snippet": snippet,
                    }
                )

        # Search classes
        for class_name, info in _codebase_cache["classes"].items():
            doc = info.get("doc") or info.get("description") or ""
            class_text = json.dumps(info)
            if query.lower() in class_text.lower():
                snippet = mcp_utils.get_snippet(doc, query)
                results["classes"].append(
                    {
                        "name": class_name,
                        "description": doc,
                        "snippet": snippet,
                    }
                )

        # Search functions
        for func_name, info in _codebase_cache["functions"].items():
            doc = info.get("doc") or info.get("description") or ""
            func_text = json.dumps(info)
            if query.lower() in func_text.lower():
                snippet = mcp_utils.get_snippet(doc, query)
                results["functions"].append(
                    {
                        "name": func_name,
                        "description": doc,
                        "signature": info.get("signature", ""),
                        "snippet": snippet,
                    }
                )

        return results

    @mcp.tool()
    async def get_function_documentation(function_name: str) -> Dict[str, Any]:
        """
        Get detailed documentation for a specific function.

        Args:
            function_name: Name of the function

        Returns:
            Dictionary with function documentation
        """
        if function_name not in _codebase_cache["functions"]:
            return {"error": f"Function {function_name} not found"}

        return _codebase_cache["functions"][function_name]

    @mcp.tool()
    async def get_class_documentation(class_name: str) -> Dict[str, Any]:
        """
        Get detailed documentation for a specific class.

        Args:
            class_name: Name of the class

        Returns:
            Dictionary with class documentation
        """
        if class_name not in _codebase_cache["classes"]:
            return {"error": f"Class {class_name} not found"}

        return _codebase_cache["classes"][class_name]



import pandas as pd
import numpy as np
import igraph as ig
import inspect

from typing import Optional


def personalized_pagerank_by_attribute(
    g: ig.Graph,
    attribute: str,
    damping: float = 0.85,
    calculate_uniform_dist: bool = True,
    additional_propagation_args: Optional[dict] = None,
) -> pd.DataFrame:
    """
    Run personalized PageRank with reset probability proportional to a vertex attribute.
    Optionally computes uniform PPR over nonzero attribute nodes.

    Parameters
    ----------
    g : igraph.Graph
        The input graph.
    attribute : str
        The vertex attribute to use for personalization.
    damping : float, optional
        Damping factor (default 0.85).
    calculate_uniform_dist : bool, optional
        If True, also compute uniform PPR over nonzero attribute nodes.
    additional_propagation_args : dict, optional
        Additional arguments to pass to igraph's personalized_pagerank. Keys must match the method's signature.

    Returns
    -------
    pd.DataFrame
        DataFrame with columns ['name', 'pagerank_by_attribute', attribute] and optionally 'pagerank_uniform'.

    Example
    -------
    >>> import igraph as ig
    >>> from scraps.utils import personalized_pagerank_by_attribute
    >>> g = ig.Graph.Full(3)
    >>> g.vs['name'] = ['A', 'B', 'C']
    >>> g.vs['score'] = [1, 0, 2]
    >>> df = personalized_pagerank_by_attribute(g, 'score')
    >>> print(df)
    """
    # Validate and extract attribute (missing/None as 0)
    attr = _ensure_nonnegative_vertex_attribute(g, attribute)

    # Validate additional_propagation_args
    if additional_propagation_args is None:
        additional_propagation_args = {}
    else:
        valid_args = set(inspect.signature(g.personalized_pagerank).parameters.keys())
        for k in additional_propagation_args:
            if k not in valid_args:
                raise ValueError(f"Invalid argument for personalized_pagerank: {k}")

    # Personalized PageRank (no normalization, igraph handles it)
    pr_attr = g.personalized_pagerank(
        reset=attr.tolist(), damping=damping, **additional_propagation_args
    )

    # Node names
    names = g.vs["name"] if "name" in g.vs.attributes() else list(range(g.vcount()))

    data = {"name": names, "pagerank_by_attribute": pr_attr, attribute: attr}

    # Uniform PPR over nonzero attribute nodes
    if calculate_uniform_dist:
        used_in_uniform = attr > 0
        n_uniform = used_in_uniform.sum()
        if n_uniform == 0:
            raise ValueError("No nonzero attribute values for uniform PPR.")
        uniform_vec = np.zeros_like(attr, dtype=float)
        uniform_vec[used_in_uniform] = 1.0 / n_uniform
        pr_uniform = g.personalized_pagerank(
            reset=uniform_vec.tolist(), damping=damping, **additional_propagation_args
        )
        data["pagerank_uniform"] = pr_uniform

    return pd.DataFrame(data)


def _ensure_nonnegative_vertex_attribute(g: ig.Graph, attribute: str):
    """
    Utility to check that a vertex attribute is present, numeric, and non-negative.
    Raises ValueError if checks fail.
    Missing or None values are treated as 0.
    Raises ValueError if attribute is missing for all vertices or all values are zero.
    """

    all_missing = all(
        (attribute not in v.attributes() or v[attribute] is None) for v in g.vs
    )
    if all_missing:
        raise ValueError(f"Vertex attribute '{attribute}' is missing for all vertices.")

    values = [
        (
            v[attribute]
            if (attribute in v.attributes() and v[attribute] is not None)
            else 0.0
        )
        for v in g.vs
    ]

    arr = np.array(values, dtype=float)

    if np.all(arr == 0):
        raise ValueError(
            f"Vertex attribute '{attribute}' is zero for all vertices; cannot use as reset vector."
        )
    if np.any(arr < 0):
        raise ValueError(f"Attribute '{attribute}' contains negative values.")

    return arr



from __future__ import annotations

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version

__version__ = "0.2.2"

try:
    __version__ = version("calicolabs-cpr")
except PackageNotFoundError:
    # package is not installed
    pass



"""
Documentation components for the Napistu MCP server.
"""

from fastmcp import FastMCP

from napistu.mcp import documentation_utils
from napistu.mcp import utils as mcp_utils
from napistu.mcp.constants import DOCUMENTATION, READMES, REPOS_WITH_ISSUES

# Global cache for documentation content
_docs_cache = {
    DOCUMENTATION.README: {},
    DOCUMENTATION.WIKI: {},
    DOCUMENTATION.ISSUES: {},
    DOCUMENTATION.PRS: {},
    DOCUMENTATION.PACKAGEDOWN: {},
}


async def initialize_components() -> bool:
    """
    Initialize documentation components.
    Loads documentation content and should be called before the server starts handling requests.

    Returns
    -------
    bool
        True if initialization is successful.
    """
    global _docs_cache
    # Load documentation from the READMES dict
    for name, url in READMES.items():
        _docs_cache[DOCUMENTATION.README][name] = (
            await documentation_utils.load_readme_content(url)
        )
    # Load issue and PR summaries with the GitHub API
    for repo in REPOS_WITH_ISSUES:
        _docs_cache[DOCUMENTATION.ISSUES][repo] = await documentation_utils.list_issues(
            repo
        )
        _docs_cache[DOCUMENTATION.PRS][repo] = (
            await documentation_utils.list_pull_requests(repo)
        )
    return True


def register_components(mcp: FastMCP):
    """
    Register documentation components with the MCP server.

    Args:
        mcp: FastMCP server instance
    """

    # Register resources
    @mcp.resource("napistu://documentation/summary")
    async def get_documentation_summary():
        """
        Get a summary of all available documentation.
        """
        return {
            "readme_files": list(_docs_cache[DOCUMENTATION.README].keys()),
            "issues": list(_docs_cache[DOCUMENTATION.ISSUES].keys()),
            "prs": list(_docs_cache[DOCUMENTATION.PRS].keys()),
            "wiki_pages": list(_docs_cache[DOCUMENTATION.WIKI].keys()),
            "packagedown_sections": list(_docs_cache[DOCUMENTATION.PACKAGEDOWN].keys()),
        }

    @mcp.resource("napistu://documentation/readme/{file_name}")
    async def get_readme_content(file_name: str):
        """
        Get the content of a specific README file.

        Args:
            file_name: Name of the README file
        """
        if file_name not in _docs_cache[DOCUMENTATION.README]:
            return {"error": f"README file {file_name} not found"}

        return {
            "content": _docs_cache[DOCUMENTATION.README][file_name],
            "format": "markdown",
        }

    @mcp.resource("napistu://documentation/issues/{repo}")
    async def get_issues(repo: str):
        """
        Get the list of issues for a given repository.
        """
        return _docs_cache[DOCUMENTATION.ISSUES].get(repo, [])

    @mcp.resource("napistu://documentation/prs/{repo}")
    async def get_prs(repo: str):
        """
        Get the list of pull requests for a given repository.
        """
        return _docs_cache[DOCUMENTATION.PRS].get(repo, [])

    @mcp.resource("napistu://documentation/issue/{repo}/{number}")
    async def get_issue_resource(repo: str, number: int):
        """
        Get a single issue by number for a given repository.
        """
        # Try cache first
        cached = next(
            (
                i
                for i in _docs_cache[DOCUMENTATION.ISSUES].get(repo, [])
                if i["number"] == number
            ),
            None,
        )
        if cached:
            return cached
        # Fallback to live fetch
        return await documentation_utils.get_issue(repo, number)

    @mcp.resource("napistu://documentation/pr/{repo}/{number}")
    async def get_pr_resource(repo: str, number: int):
        """
        Get a single pull request by number for a given repository.
        """
        # Try cache first
        cached = next(
            (
                pr
                for pr in _docs_cache[DOCUMENTATION.PRS].get(repo, [])
                if pr["number"] == number
            ),
            None,
        )
        if cached:
            return cached
        # Fallback to live fetch
        return await documentation_utils.get_issue(repo, number)

    # Register tools
    @mcp.tool("search_documentation")
    async def search_documentation(query: str):
        """
        Search all documentation for a specific query.

        Args:
            query: Search term

        Returns:
            Dictionary with search results organized by documentation type
        """
        results = {
            DOCUMENTATION.README: [],
            DOCUMENTATION.WIKI: [],
            DOCUMENTATION.ISSUES: [],
            DOCUMENTATION.PRS: [],
            DOCUMENTATION.PACKAGEDOWN: [],
        }
        # Simple text search
        for readme_name, content in _docs_cache[DOCUMENTATION.README].items():
            if query.lower() in content.lower():
                results[DOCUMENTATION.README].append(
                    {
                        "name": readme_name,
                        "snippet": mcp_utils.get_snippet(content, query),
                    }
                )
        return results



"""Module to contain all constants for CPR"""

from __future__ import annotations

import libsbml

from types import SimpleNamespace
import pandas as pd


PACKAGE_DEFS = SimpleNamespace(
    NAPISTU="napistu",
    GITHUB_OWNER="napistu",
    GITHUB_PROJECT_REPO="napistu",
    GITHUB_NAPISTU_PY="napistu-py",
    GITHUB_NAPISTU_R="napistu-r",
    TUTORIALS_URL="https://github.com/napistu/napistu/wiki",
    # User-facing functionality should use a user-defined directory but
    # for convenience, we provide a default cache directory for dev-facing
    # workflows
    CACHE_DIR="napistu_data",
)

PROTEINATLAS_SUBCELL_LOC_URL = (
    "https://www.proteinatlas.org/download/tsv/subcellular_location.tsv.zip"
)

# GTEx
GTEX_RNASEQ_EXPRESSION_URL = "https://storage.googleapis.com/adult-gtex/bulk-gex/v8/rna-seq/GTEx_Analysis_2017-06-05_v8_RNASeQCv1.1.9_gene_median_tpm.gct.gz"

# Gencode
GENCODE_URL = "https://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_human/release_26/gencode.v26.transcripts.fa.gz"

FILE_EXT_ZIP = "zip"
FILE_EXT_GZ = "gz"

# SBML_dfs

SBML_DFS = SimpleNamespace(
    COMPARTMENTS="compartments",
    SPECIES="species",
    COMPARTMENTALIZED_SPECIES="compartmentalized_species",
    REACTIONS="reactions",
    REACTION_SPECIES="reaction_species",
    SPECIES_DATA="species_data",
    REACTIONS_DATA="reactions_data",
    C_ID="c_id",
    C_NAME="c_name",
    C_IDENTIFIERS="c_Identifiers",
    C_SOURCE="c_Source",
    S_ID="s_id",
    S_NAME="s_name",
    S_IDENTIFIERS="s_Identifiers",
    S_SOURCE="s_Source",
    SC_ID="sc_id",
    SC_NAME="sc_name",
    SC_SOURCE="sc_Source",
    R_ID="r_id",
    R_NAME="r_name",
    R_IDENTIFIERS="r_Identifiers",
    R_SOURCE="r_Source",
    R_ISREVERSIBLE="r_isreversible",
    RSC_ID="rsc_id",
    STOICHIOMETRY="stoichiometry",
    SBO_TERM="sbo_term",
)

SBML_DFS_SCHEMA = SimpleNamespace(
    SCHEMA={
        SBML_DFS.COMPARTMENTS: {
            "pk": SBML_DFS.C_ID,
            "label": SBML_DFS.C_NAME,
            "id": SBML_DFS.C_IDENTIFIERS,
            "source": SBML_DFS.C_SOURCE,
            "vars": [SBML_DFS.C_NAME, SBML_DFS.C_IDENTIFIERS, SBML_DFS.C_SOURCE],
        },
        SBML_DFS.SPECIES: {
            "pk": SBML_DFS.S_ID,
            "label": SBML_DFS.S_NAME,
            "id": SBML_DFS.S_IDENTIFIERS,
            "source": SBML_DFS.S_SOURCE,
            "vars": [SBML_DFS.S_NAME, SBML_DFS.S_IDENTIFIERS, SBML_DFS.S_SOURCE],
        },
        SBML_DFS.COMPARTMENTALIZED_SPECIES: {
            "pk": SBML_DFS.SC_ID,
            "label": SBML_DFS.SC_NAME,
            "fk": [SBML_DFS.S_ID, SBML_DFS.C_ID],
            "source": SBML_DFS.SC_SOURCE,
            "vars": [
                SBML_DFS.SC_NAME,
                SBML_DFS.S_ID,
                SBML_DFS.C_ID,
                SBML_DFS.SC_SOURCE,
            ],
        },
        SBML_DFS.REACTIONS: {
            "pk": SBML_DFS.R_ID,
            "label": SBML_DFS.R_NAME,
            "id": SBML_DFS.R_IDENTIFIERS,
            "source": SBML_DFS.R_SOURCE,
            "vars": [
                SBML_DFS.R_NAME,
                SBML_DFS.R_IDENTIFIERS,
                SBML_DFS.R_SOURCE,
                SBML_DFS.R_ISREVERSIBLE,
            ],
        },
        SBML_DFS.REACTION_SPECIES: {
            "pk": SBML_DFS.RSC_ID,
            "fk": [SBML_DFS.R_ID, SBML_DFS.SC_ID],
            "vars": [
                SBML_DFS.R_ID,
                SBML_DFS.SC_ID,
                SBML_DFS.STOICHIOMETRY,
                SBML_DFS.SBO_TERM,
            ],
        },
    },
    REQUIRED_ENTITIES={
        SBML_DFS.COMPARTMENTS,
        SBML_DFS.SPECIES,
        SBML_DFS.COMPARTMENTALIZED_SPECIES,
        SBML_DFS.REACTIONS,
        SBML_DFS.REACTION_SPECIES,
    },
    OPTIONAL_ENTITIES={
        SBML_DFS.SPECIES_DATA,
        SBML_DFS.REACTIONS_DATA,
    },
)

ENTITIES_W_DATA = {SBML_DFS.SPECIES, SBML_DFS.REACTIONS}

REQUIRED_REACTION_FROMEDGELIST_COLUMNS = [
    "sc_id_up",
    "sc_id_down",
    "sbo_term",
    "r_name",
    "r_Identifiers",
    "r_isreversible",
]

CPR_STANDARD_OUTPUTS = SimpleNamespace(
    SPECIES_IDENTIFIERS="species_identifiers.tsv",
    SPECIES="species.json",
    REACTIONS="reactions.json",
    REACTION_SPECIES="reaction_species.json",
    COMPARTMENTS="compartments.json",
    COMPARTMENTALIZED_SPECIES="compartmentalized_species.json",
)

INTERACTION_EDGELIST_EXPECTED_VARS = {
    "upstream_name",
    "downstream_name",
    "upstream_compartment",
    "downstream_compartment",
    "r_name",
    "sbo_term",
    "r_Identifiers",
    "r_isreversible",
}

BQB_PRIORITIES = pd.DataFrame(
    [{"bqb": "BQB_IS", "bqb_rank": 1}, {"bqb": "BQB_HAS_PART", "bqb_rank": 2}]
)

ONTOLOGY_PRIORITIES = pd.DataFrame(
    [
        {"ontology": "reactome", "ontology_rank": 1},
        {"ontology": "ensembl_gene", "ontology_rank": 2},
        {"ontology": "chebi", "ontology_rank": 3},
        {"ontology": "uniprot", "ontology_rank": 4},
        {"ontology": "go", "ontology_rank": 5},
    ]
)

# SBML
# Biological qualifiers
# Biomodels qualifiers
BQB = SimpleNamespace(
    IS="BQB_IS",
    HAS_PART="BQB_HAS_PART",
    IS_PART_OF="BQB_IS_PART_OF",
    IS_VERSION_OF="BQB_IS_VERSION_OF",
    HAS_VERSION="BQB_HAS_VERSION",
    IS_HOMOLOG_TO="BQB_IS_HOMOLOG_TO",
    IS_DESCRIBED_BY="BQB_IS_DESCRIBED_BY",
    IS_ENCODED_BY="BQB_IS_ENCODED_BY",
    ENCODES="BQB_ENCODES",
    OCCURS_IN="BQB_OCCURS_IN",
    HAS_PROPERTY="BQB_HAS_PROPERTY",
    IS_PROPERTY_OF="BQB_IS_PROPERTY_OF",
    HAS_TAXON="BQB_HAS_TAXON",
    UNKNOWN="BQB_UNKNOWN",
)

# molecules are distinctly defined by these BQB terms
BQB_DEFINING_ATTRS = ["BQB_IS", "IS_HOMOLOG_TO"]

# a looser convention which will aggregate genes, transcripts, and proteins
# if they are linked with the appropriate bioqualifiers
BQB_DEFINING_ATTRS_LOOSE = [
    "BQB_IS",
    "IS_HOMOLOG_TO",
    "BQB_IS_ENCODED_BY",
    "BQB_ENCODES",
]

# identifiers
IDENTIFIERS = SimpleNamespace(
    ONTOLOGY="ontology", IDENTIFIER="identifier", BQB="bqb", URL="url"
)

SPECIES_IDENTIFIERS_REQUIRED_VARS = {
    SBML_DFS.S_ID,
    IDENTIFIERS.ONTOLOGY,
    IDENTIFIERS.IDENTIFIER,
    IDENTIFIERS.BQB,
    SBML_DFS.S_NAME,
}

BIOLOGICAL_QUALIFIERS = [
    "BQB_IS",
    "BQB_HAS_PART",
    "BQB_IS_PART_OF",
    "BQB_IS_VERSION_OF",
    "BQB_HAS_VERSION",
    "BQB_IS_HOMOLOG_TO",
    "BQB_IS_DESCRIBED_BY",
    "BQB_IS_ENCODED_BY",
    "BQB_ENCODES",
    "BQB_OCCURS_IN",
    "BQB_HAS_PROPERTY",
    "BQB_IS_PROPERTY_OF",
    "BQB_HAS_TAXON",
    "BQB_UNKNOWN",
]


def get_biological_qualifier_codes():
    bio_qualifier_codes = {getattr(libsbml, bqb): bqb for bqb in BIOLOGICAL_QUALIFIERS}

    return bio_qualifier_codes


BIOLOGICAL_QUALIFIER_CODES = get_biological_qualifier_codes()

# Systems biology ontology
SBOTERM_NAMES = SimpleNamespace(
    REACTANT="reactant",
    PRODUCT="product",
    CATALYST="catalyst",
    INHIBITOR="inhibitor",
    STIMULATOR="stimulator",
    MODIFIER="modifier",
    INTERACTOR="interactor",
)

MINI_SBO_TO_NAME = {
    "SBO:0000010": SBOTERM_NAMES.REACTANT,
    "SBO:0000011": SBOTERM_NAMES.PRODUCT,
    "SBO:0000013": SBOTERM_NAMES.CATALYST,
    "SBO:0000020": SBOTERM_NAMES.INHIBITOR,
    "SBO:0000459": SBOTERM_NAMES.STIMULATOR,
    "SBO:0000019": SBOTERM_NAMES.MODIFIER,
    "SBO:0000336": SBOTERM_NAMES.INTERACTOR,
}

MINI_SBO_FROM_NAME = {
    SBOTERM_NAMES.REACTANT: "SBO:0000010",
    SBOTERM_NAMES.PRODUCT: "SBO:0000011",
    SBOTERM_NAMES.CATALYST: "SBO:0000013",
    SBOTERM_NAMES.INHIBITOR: "SBO:0000020",
    SBOTERM_NAMES.STIMULATOR: "SBO:0000459",
    SBOTERM_NAMES.MODIFIER: "SBO:0000019",  # parent category of inhibitor and stimulator (i.e., activator)
    SBOTERM_NAMES.INTERACTOR: "SBO:0000336",  # entity participating in a physical or functional interaction
}

SBO_MODIFIER_NAMES = {
    SBOTERM_NAMES.INHIBITOR,
    SBOTERM_NAMES.STIMULATOR,
    SBOTERM_NAMES.MODIFIER,
}

MINI_SBO_NAME_TO_POLARITY = {
    SBOTERM_NAMES.REACTANT: "activation",
    SBOTERM_NAMES.PRODUCT: "activation",
    SBOTERM_NAMES.CATALYST: "activation",
    SBOTERM_NAMES.INHIBITOR: "inhibition",
    SBOTERM_NAMES.STIMULATOR: "activation",
    SBOTERM_NAMES.MODIFIER: "ambiguous",
    SBOTERM_NAMES.INTERACTOR: "ambiguous",
}

# how does changing a reactions' membership
# affect whether a reaction can occur
# for example, if I remove any substrate a reaction won't occur
# but I would have to remove all catalysts for it to not occur
SBO_NAME_TO_ROLE = {
    SBOTERM_NAMES.REACTANT: "DEFINING",
    SBOTERM_NAMES.PRODUCT: "DEFINING",
    SBOTERM_NAMES.INTERACTOR: "DEFINING",
    SBOTERM_NAMES.CATALYST: "REQUIRED",
    SBOTERM_NAMES.INHIBITOR: "OPTIONAL",
    SBOTERM_NAMES.STIMULATOR: "OPTIONAL",
    SBOTERM_NAMES.MODIFIER: "OPTIONAL",
}

# see also https://github.com/calico/netcontextr/blob/main/R/reactionTrimmingFunctions.R
VALID_SBO_ROLES = (
    # there is a direct correspondence between the set of defining entries and the identity of a reaction
    # e.g., the stoichiometery of a metabolic reaction or the members of a protein-protein interaction
    "DEFINING",
    # 1+ entries are needed if entries were initially defined. i.e., reactions which require a catalyst
    # would no longer exist if the catalyst was removed, but many reactions do not require a catalyst.
    "REQUIRED",
    # 0+ entries. optional species can be added or removed to a reaction without changing its identity
    "OPTIONAL",
)

# required variables for the edgelist formats used by mechanism_matching
CPR_EDGELIST = SimpleNamespace(
    S_ID_UPSTREAM="s_id_upstream",
    S_ID_DOWNSTREAM="s_id_downstream",
    SC_ID_UPSTREAM="sc_id_upstream",
    SC_ID_DOWNSTREAM="sc_id_downstream",
    IDENTIFIER_UPSTREAM="identifier_upstream",
    IDENTIFIER_DOWNSTREAM="identifier_downstream",
    S_NAME_UPSTREAM="s_name_upstream",
    S_NAME_DOWNSTREAM="s_name_downstream",
    SC_ID_ORIGIN="sc_id_origin",
    SC_ID_DEST="sc_id_dest",
)

IDENTIFIER_EDGELIST_REQ_VARS = {
    CPR_EDGELIST.IDENTIFIER_UPSTREAM,
    CPR_EDGELIST.IDENTIFIER_DOWNSTREAM,
}

CPR_EDGELIST_REQ_VARS = {
    CPR_EDGELIST.S_ID_UPSTREAM,
    CPR_EDGELIST.S_ID_DOWNSTREAM,
    CPR_EDGELIST.SC_ID_UPSTREAM,
    CPR_EDGELIST.SC_ID_DOWNSTREAM,
}

CPR_PATH_REQ_VARS = {CPR_EDGELIST.SC_ID_ORIGIN, CPR_EDGELIST.SC_ID_DEST}

FEATURE_ID_VAR_DEFAULT = "feature_id"

RESOLVE_MATCHES_AGGREGATORS = SimpleNamespace(
    WEIGHTED_MEAN="weighted_mean", MEAN="mean", FIRST="first", MAX="max"
)

RESOLVE_MATCHES_TMP_WEIGHT_COL = "__tmp_weight_for_aggregation__"

# specifying weighting schemes schema

DEFAULT_WT_TRANS = "identity"

DEFINED_WEIGHT_TRANSFORMATION = {
    DEFAULT_WT_TRANS: "_wt_transformation_identity",
    "string": "_wt_transformation_string",
    "string_inv": "_wt_transformation_string_inv",
}

SCORE_CALIBRATION_POINTS_DICT = {
    "weights": {"strong": 3, "good": 7, "okay": 20, "weak": 40},
    "string_wt": {"strong": 950, "good": 400, "okay": 230, "weak": 150},
}

SOURCE_VARS_DICT = {"string_wt": 10}

# source
SOURCE_SPEC = SimpleNamespace(
    PATHWAY_ID="pathway_id",
    MODEL="model",
    SOURCE="source",
    SPECIES="species",
    NAME="name",
    ENTRY="entry",
    N_COLLAPSED_PATHWAYS="n_collapsed_pathways",
    INDEX_NAME="entry",
    FILE="file",
    DATE="date",
)

EXPECTED_PW_INDEX_COLUMNS = {
    SOURCE_SPEC.FILE,
    SOURCE_SPEC.PATHWAY_ID,
    SOURCE_SPEC.SOURCE,
    SOURCE_SPEC.SPECIES,
    SOURCE_SPEC.NAME,
    SOURCE_SPEC.DATE,
}

# rules for specific ontologies

ONTOLOGIES = SimpleNamespace(
    CHEBI="chebi",
    ENSEMBL_GENE="ensembl_gene",
    ENSEMBL_TRANSCRIPT="ensembl_transcript",
    ENSEMBL_PROTEIN="ensembl_protein",
    GENE_NAME="gene_name",
    GO="go",
    MIRBASE="mirbase",
    NCBI_ENTREZ_GENE="ncbi_entrez_gene",
    PHAROS="pharos",
    REACTOME="reactome",
    SYMBOL="symbol",
    UNIPROT="uniprot",
)

ONTOLOGIES_LIST = list(ONTOLOGIES.__dict__.values())

CHARACTERISTIC_COMPLEX_ONTOLOGIES = [
    ONTOLOGIES.ENSEMBL_GENE,
    ONTOLOGIES.NCBI_ENTREZ_GENE,
    ONTOLOGIES.MIRBASE,
]

ONTOLOGY_ALIASES = SimpleNamespace(NCBI_ENTREZ_GENE={"ncbigene", "ncbi_gene"})

ENSEMBL_MOLECULE_TYPES_TO_ONTOLOGY = {
    "G": ONTOLOGIES.ENSEMBL_GENE,
    "T": ONTOLOGIES.ENSEMBL_TRANSCRIPT,
    "P": ONTOLOGIES.ENSEMBL_PROTEIN,
}

ENSEMBL_MOLECULE_TYPES_FROM_ONTOLOGY = {
    ONTOLOGIES.ENSEMBL_GENE: "G",
    ONTOLOGIES.ENSEMBL_TRANSCRIPT: "T",
    ONTOLOGIES.ENSEMBL_PROTEIN: "P",
}

ENSEMBL_SPECIES_FROM_CODE = {"MUS": "Mus musculus"}

ENSEMBL_SPECIES_TO_CODE = {"Mus musculus": "MUS"}

ENSEMBL_PREFIX_TO_ONTOLOGY = {
    "ENSG": ONTOLOGIES.ENSEMBL_GENE,
    "ENST": ONTOLOGIES.ENSEMBL_TRANSCRIPT,
    "ENSP": ONTOLOGIES.ENSEMBL_PROTEIN,
}

COMPARTMENTS = {
    "NUCLEOPLASM": "nucleoplasm",
    "CYTOPLASM": "cytoplasm",
    "CELLULAR_COMPONENT": "cellular_component",
    "CYTOSOL": "cytosol",
    "MITOCHONDRIA": "mitochondria",
    "MITOMEMBRANE": "mitochondrial membrane",
    "INNERMITOCHONDRIA": "inner mitochondria",
    "MITOMATRIX": "mitochondrial matrix",
    "ENDOPLASMICRETICULUM": "endoplasmic reticulum",
    "ERMEMBRANE": "endoplasmic reticulum membrane",
    "ERLUMEN": "endoplasmic reticulum lumen",
    "GOLGIAPPARATUS": "golgi apparatus",
    "GOLGIMEMBRANE": "golgi membrane",
    "NUCLEUS": "nucleus",
    "NUCLEARLUMEN": "nuclear lumen",
    "NUCLEOLUS": "nucleolus",
    "LYSOSOME": "lysosome",
    "PEROXISOME": "peroxisome",
    "EXTRACELLULAR": "extracellular",
}

COMPARTMENT_ALIASES = {
    "NUCLEOPLASM": ["nucleoplasm", "Nucleoplasm"],
    "CYTOPLASM": ["cytoplasm", "Cytoplasm"],
    "CELLULAR_COMPONENT": ["cellular_component", "Cellular_component"],
    "CYTOSOL": ["cytosol", "Cytosol"],
    "MITOCHONDRIA": ["mitochondria", "Mitochondria"],
    "MITOMEMBRANE": ["mitochondrial membrane", "Mitochondrial membrane"],
    "INNERMITOCHONDRIA": [
        "inner mitochondria",
        "Inner mitochondria",
        "inner mitochondrial compartment",
    ],
    "MITOMATRIX": [
        "mitochondrial matrix",
        "Mitochondrial matrix",
        "mitochondrial lumen",
        "Mitochondrial lumen",
    ],
    "ENDOPLASMICRETICULUM": ["endoplasmic reticulum", "Endoplasmic reticulum"],
    "ERMEMBRANE": ["endoplasmic reticulum membrane", "Endoplasmic reticulum membrane"],
    "ERLUMEN": ["endoplasmic reticulum lumen", "Endoplasmic reticulum lumen"],
    "GOLGIAPPARATUS": ["golgi apparatus", "Golgi apparatus"],
    "GOLGIMEMBRANE": ["Golgi membrane", "golgi membrane"],
    "NUCLEUS": ["nucleus", "Nucleus"],
    "NUCLEARLUMEN": ["nuclear lumen", "Nuclear lumen"],
    "NUCLEOLUS": ["nucleolus", "Nucleolus"],
    "LYSOSOME": ["lysosome", "Lysosome"],
    "PEROXISOME": ["peroxisome", "Peroxisome", "peroxisome/glyoxysome"],
    "EXTRACELLULAR": [
        "extracellular",
        "Extracellular",
        "extracellular space",
        "Extracellular space",
    ],
}

COMPARTMENTS_GO_TERMS = {
    "NUCLEOPLASM": "GO:0005654",
    "CELLULAR_COMPONENT": "GO:0005575",
    "CYTOPLASM": "GO:0005737",
    "CYTOSOL": "GO:0005829",
    "MITOCHONDRIA": "GO:0005739",
    "MITOMEMBRANE": "GO:0031966",
    "INNERMITOCHONDRIA": "GO:0005743",
    "MITOMATRIX": "GO:0005759",
    "ENDOPLASMICRETICULUM": "GO:0005783",
    "ERMEMBRANE": "GO:0005789",
    "ERLUMEN": "GO:0005788",
    "GOLGIAPPARATUS": "GO:0005794",
    "GOLGIMEMBRANE": "GO:0000139",
    "NUCLEUS": "GO:0005634",
    "NUCLEARLUMEN": "GO:0031981",
    "NUCLEOLUS": "GO:0005730",
    "LYSOSOME": "GO:0005764",
    "PEROXISOME": "GO:0005777",
    "EXTRACELLULAR": "GO:0005615",
}



from __future__ import annotations

import datetime
import logging
import os
from typing import Iterable

import pandas as pd
from napistu import indices
from napistu import sbml_dfs_core
from napistu import utils
from napistu.consensus import construct_sbml_dfs_dict
from napistu.ingestion import sbml
from napistu.ingestion.constants import BIGG_MODEL_FIELD_SPECIES
from napistu.ingestion.constants import BIGG_MODEL_FIELD_URL
from napistu.ingestion.constants import BIGG_MODEL_KEYS
from napistu.ingestion.constants import BIGG_MODEL_URLS
from napistu.ingestion.constants import BIGG_RECON3D_FIELD_ANNOTATION
from napistu.ingestion.constants import SPECIES_FULL_NAME_HUMAN
from napistu.ingestion.constants import SPECIES_FULL_NAME_MOUSE
from napistu.ingestion.constants import SPECIES_FULL_NAME_YEAST
from fs import open_fs

logger = logging.getLogger(__name__)


def bigg_sbml_download(bg_pathway_root: str, overwrite: bool = False) -> None:
    """
    BiGG SBML Download

    Download SBML models from BiGG. Currently just the human Recon3D model

    Parameters:
    bg_pathway_root (str): Paths to a directory where a \"sbml\" directory should be created.
    overwrite (bool): Overwrite an existing output directory.

    Returns:
    None

    """
    utils.initialize_dir(bg_pathway_root, overwrite)

    bigg_models = {
        BIGG_MODEL_KEYS[SPECIES_FULL_NAME_HUMAN]: {
            BIGG_MODEL_FIELD_URL: BIGG_MODEL_URLS[SPECIES_FULL_NAME_HUMAN],
            BIGG_MODEL_FIELD_SPECIES: SPECIES_FULL_NAME_HUMAN,
        },
        BIGG_MODEL_KEYS[SPECIES_FULL_NAME_MOUSE]: {
            BIGG_MODEL_FIELD_URL: BIGG_MODEL_URLS[SPECIES_FULL_NAME_MOUSE],
            BIGG_MODEL_FIELD_SPECIES: SPECIES_FULL_NAME_MOUSE,
        },
        BIGG_MODEL_KEYS[SPECIES_FULL_NAME_YEAST]: {
            BIGG_MODEL_FIELD_URL: BIGG_MODEL_URLS[SPECIES_FULL_NAME_YEAST],
            BIGG_MODEL_FIELD_SPECIES: SPECIES_FULL_NAME_YEAST,
        },
    }
    bigg_models_df = pd.DataFrame(bigg_models).T
    bigg_models_df["sbml_path"] = [
        os.path.join(bg_pathway_root, k) + ".sbml"
        for k in bigg_models_df.index.tolist()
    ]
    bigg_models_df["file"] = [os.path.basename(x) for x in bigg_models_df["sbml_path"]]

    # add other attributes which will be used in the pw_index
    bigg_models_df["date"] = datetime.date.today().strftime("%Y%m%d")
    bigg_models_df.index = bigg_models_df.index.rename("pathway_id")
    bigg_models_df = bigg_models_df.reset_index()
    bigg_models_df["name"] = bigg_models_df["pathway_id"]
    bigg_models_df = bigg_models_df.assign(source="BiGG")

    with open_fs(bg_pathway_root, create=True) as bg_fs:
        for _, row in bigg_models_df.iterrows():
            with bg_fs.open(row["file"], "wb") as f:
                utils.download_wget(row["url"], f)  # type: ignore

        pw_index = bigg_models_df[
            ["file", "source", "species", "pathway_id", "name", "date"]
        ]

        # save index to sbml dir
        with bg_fs.open("pw_index.tsv", "wb") as f:
            pw_index.to_csv(f, sep="\t", index=False)

    return None


def annotate_recon(raw_model_path: str, annotated_model_path: str) -> None:
    """Annotate Recon3D
    Add compartment annotations to Recon3D so it can be merged with other pathways
    """
    logger.warning(
        "add_sbml_annotations is deprecated and maybe removed in a future version of rcpr; "
        "we are now adding these annotation during ingestion by sbml.sbml_df_from_sbml() rather "
        "than directly appending them to the raw .sbml"
    )
    recon_3d_annotations = pd.DataFrame(BIGG_RECON3D_FIELD_ANNOTATION)
    sbml_model = sbml.SBML(raw_model_path)
    sbml.add_sbml_annotations(
        sbml_model, recon_3d_annotations, save_path=annotated_model_path
    )

    return None


def construct_bigg_consensus(
    pw_index_inp: str | indices.PWIndex,
    species: str | Iterable[str] | None = None,
    outdir: str | None = None,
) -> sbml_dfs_core.SBML_dfs:
    """Constructs a BiGG SBML DFs Pathway Representation

    Attention: curently this does work only for a singly model. Integraiton of multiple
    models is not supported yet in BiGG.

    Args:
        pw_index_inp (str | indices.PWIndex): PWIndex or uri pointing to PWIndex
        species (str | Iterable[str] | None): one or more species to filter by. Default: no filtering
        outdir (str | None, optional): output directory used to cache results. Defaults to None.

    Returns:
        sbml_dfs_core.SBML_dfs: A consensus SBML
    """
    if isinstance(pw_index_inp, str):
        pw_index = indices.adapt_pw_index(pw_index_inp, species=species, outdir=outdir)
    elif isinstance(pw_index_inp, indices.PWIndex):
        pw_index = pw_index_inp
    else:
        raise ValueError("pw_index_inp needs to be a PWIndex or a str to a location.")
    if outdir is not None:
        construct_sbml_dfs_dict_fkt = utils.pickle_cache(
            os.path.join(outdir, "model_pool.pkl")
        )(construct_sbml_dfs_dict)
    else:
        construct_sbml_dfs_dict_fkt = construct_sbml_dfs_dict

    sbml_dfs_dict = construct_sbml_dfs_dict_fkt(pw_index)
    if len(sbml_dfs_dict) > 1:
        raise NotImplementedError("Merging of models not implemented yet for BiGG")

    # In Bigg there should be only one model
    model = list(sbml_dfs_dict.values())[0]
    # fix missing compartimentalization
    model = sbml_dfs_core.infer_uncompartmentalized_species_location(model)
    model = sbml_dfs_core.name_compartmentalized_species(model)
    model.validate()
    return model
